import socket
import time

import pytest
from six.moves import queue

from nameko import config
from nameko.constants import WEB_SERVER_CONFIG_KEY
from nameko.extensions import DependencyProvider
from nameko.rpc import ServiceRpc, rpc
from nameko.standalone.rpc import ServiceRpcClient
from nameko.testing import rabbit
from nameko.testing.utils import get_rabbit_connections
from nameko.web.handlers import http
from nameko.web.server import parse_address
from nameko.web.websocket import rpc as wsrpc


pytest_plugins = "pytester"


@pytest.fixture
def plugin_options(request):
    """ Get the options pytest may have been invoked with so we can pass
    them into subprocess pytests created by the pytester plugin.
    """
    options = (
        '--rabbit-amqp-uri',
        '--rabbit-api-uri'
    )

    args = [
        "{}={}".format(opt, request.config.getoption(opt)) for opt in options
    ]
    return args


class TestOptions(object):

    def test_options(self, testdir):

        options = (
            ("--amqp-uri", 'amqp://localhost:5672/vhost'),
            ("--rabbit-api-uri", 'http://localhost:15672'),
            ("--amqp-ssl-port", '1234'),
        )

        testdir.makepyfile(
            """
            import re

            def test_option(request):
                assert request.config.getoption('RABBIT_AMQP_URI') == (
                    'amqp://localhost:5672/vhost'
                )
                assert request.config.getoption('RABBIT_API_URI') == (
                    'http://localhost:15672'
                )
                assert request.config.getoption('AMQP_SSL_PORT') == '1234'
            """
        )
        args = []
        for option, value in options:
            args.extend((option, value))
        result = testdir.runpytest(*args)
        assert result.ret == 0

    def test_ssl_options(self, testdir):

        options = (
            ('certfile', 'path/cert.pem'),  # override default
            ('string', 'string'),
            ('number', 1),
            ('list', '[1, 2, 3]'),
            ('map', '{"foo": "bar"}'),
            ('lookup', '!!python/name:ssl.CERT_REQUIRED'),
        )

        testdir.makepyfile(
            """
            import re
            import ssl

            import pytest

            from nameko import config

            @pytest.mark.usefixtures("rabbit_ssl_config")
            def test_ssl_options(request):
                assert request.config.getoption('AMQP_SSL_OPTIONS') == [
                    # defaults
                    ('ca_certs', 'certs/cacert.pem'),
                    ('certfile', 'certs/clientcert.pem'),
                    ('keyfile', 'certs/clientkey.pem'),
                    ('cert_reqs', ssl.CERT_REQUIRED),
                    # additions
                    ('certfile', 'path/cert.pem'),
                    ('string', 'string'),
                    ('number', 1),
                    ('list', [1, 2, 3]),
                    ('map', {'foo': 'bar'}),
                    ('lookup', ssl.CERT_REQUIRED),
                    ('keyonly', True),
                ]

                expected_ssl_options = {
                    'certfile': 'path/cert.pem',  # default overridden
                    'ca_certs': 'certs/cacert.pem',
                    'keyfile': 'certs/clientkey.pem',
                    'cert_reqs': ssl.CERT_REQUIRED,
                    'string': 'string',
                    'number': 1,
                    'list': [1, 2, 3],
                    'map': {'foo': 'bar'},
                    'lookup': ssl.CERT_REQUIRED,
                    'keyonly': True,
                }
                assert config['AMQP_SSL'] == expected_ssl_options
            """
        )
        args = []
        for key, value in options:
            args.extend(['--amqp-ssl-option', "{}={}".format(key, value)])
        # key only case
        args.extend(['--amqp-ssl-option', "keyonly"])

        result = testdir.runpytest(*args)
        assert result.ret == 0


def test_empty_config(empty_config):
    assert config == {}


def test_rabbit_manager(rabbit_manager):
    assert isinstance(rabbit_manager, rabbit.Client)
    assert "/" in [vhost['name'] for vhost in rabbit_manager.get_all_vhosts()]


def test_amqp_uri(testdir):

    amqp_uri = "amqp://user:pass@host:5672/vhost"

    testdir.makeconftest(
        """
        import pytest
        from nameko import config

        @pytest.yield_fixture
        def rabbit_config():
            with config.patch(dict(AMQP_URI="{}")):
                yield
        """.format(amqp_uri)
    )

    testdir.makepyfile(
        """
        import re

        def test_amqp_uri(amqp_uri):
            assert amqp_uri == '{}'
        """.format(amqp_uri)
    )
    result = testdir.runpytest(
        "--amqp-uri", amqp_uri
    )
    assert result.ret == 0


class TestGetMessageFromQueue(object):

    @pytest.fixture
    def queue_name(self):
        return "queue"

    @pytest.fixture
    def publish_message(
        self, rabbit_manager, rabbit_config, get_vhost, queue_name
    ):
        vhost = get_vhost(config['AMQP_URI'])
        rabbit_manager.create_queue(vhost, queue_name, durable=True)

        def publish(payload, **properties):
            rabbit_manager.publish(
                vhost, "amq.default", queue_name, payload, properties
            )

        return publish

    @pytest.mark.usefixtures("rabbit_config")
    def test_get_message(
        self, publish_message, get_message_from_queue, queue_name,
        rabbit_manager, get_vhost
    ):
        payload = "payload"
        publish_message(payload)

        message = get_message_from_queue(queue_name)
        assert message.payload == payload

        vhost = get_vhost(config['AMQP_URI'])
        assert rabbit_manager.get_queue(vhost, queue_name)['messages'] == 0

    @pytest.mark.usefixtures("rabbit_config")
    def test_requeue(
        self, publish_message, get_message_from_queue, queue_name,
        rabbit_manager, get_vhost
    ):
        payload = "payload"
        publish_message(payload)

        message = get_message_from_queue(queue_name, ack=False)
        assert message.payload == payload

        time.sleep(1)  # TODO: use retry decorator rather than sleep
        vhost = get_vhost(config['AMQP_URI'])
        assert rabbit_manager.get_queue(vhost, queue_name)['messages'] == 1

    def test_non_blocking(
        self, publish_message, get_message_from_queue, queue_name
    ):
        # no message published; raises immediately
        with pytest.raises(queue.Empty):
            get_message_from_queue(queue_name, block=False)

    def test_timeout(
        self, publish_message, get_message_from_queue, queue_name
    ):
        # no message published; raises after timeout
        with pytest.raises(queue.Empty):
            get_message_from_queue(queue_name, timeout=0.01)

    def test_accept(
        self, publish_message, get_message_from_queue, queue_name
    ):
        payload = "payload"
        content_type = "application/x-special"
        publish_message(payload, content_type=content_type)

        message = get_message_from_queue(queue_name, accept=content_type)
        assert message.properties['content_type'] == content_type
        assert message.payload == payload


@pytest.mark.usefixtures("rabbit_config")
def test_container_factory(
    testdir, get_vhost, rabbit_manager, plugin_options
):

    testdir.makepyfile(
        """
        import pytest

        from nameko.rpc import rpc
        from nameko.standalone.rpc import ServiceRpcClient

        class ServiceX(object):
            name = "x"

            @rpc
            def method(self):
                return "OK"

        @pytest.mark.usefixtures('rabbit_config')
        def test_container_factory(container_factory):
            container = container_factory(ServiceX)
            container.start()

            with ServiceRpcClient("x") as client:
                assert client.method() == "OK"
        """
    )
    result = testdir.runpytest(*plugin_options)
    assert result.ret == 0

    vhost = get_vhost(config['AMQP_URI'])
    assert get_rabbit_connections(vhost, rabbit_manager) == []


def test_container_factory_with_custom_container_cls(testdir, plugin_options):

    testdir.makepyfile(container_module="""
        from nameko.containers import ServiceContainer

        class ServiceContainerX(ServiceContainer):
            pass
    """)

    testdir.makepyfile(
        """
        import pytest

        from nameko import config
        from nameko.rpc import rpc
        from nameko.standalone.rpc import ServiceRpcClient

        from container_module import ServiceContainerX

        class ServiceX(object):
            name = "x"

            @rpc
            def method(self):
                return "OK"

        @pytest.mark.usefixtures('rabbit_config')
        def test_container_factory(container_factory):
            with config.patch({
                'SERVICE_CONTAINER_CLS': "container_module.ServiceContainerX"
            }):
                container = container_factory(ServiceX)
                container.start()

                assert isinstance(container, ServiceContainerX)

            with ServiceRpcClient("x") as client:
                assert client.method() == "OK"
        """
    )
    result = testdir.runpytest(*plugin_options)
    assert result.ret == 0


def test_container_factory_with_config_patch_argument(testdir):

    testdir.makepyfile(
        """
        import nameko
        import pytest

        from nameko.testing.services import dummy

        class ServiceX(object):
            name = "x"

            @dummy
            def method(self):
                return "OK"

        def test_container_factory_applies_config_patch_(container_factory):

            assert "FOO" not in nameko.config

            container = container_factory(ServiceX, {"FOO": 2})
            container.start()

            assert nameko.config["FOO"] == 2

        def test_container_factory_rolls_back_config_patch():

            assert "FOO" not in nameko.config
        """
    )
    result = testdir.runpytest()
    assert result.ret == 0


@pytest.mark.usefixtures("rabbit_config")
def test_runner_factory(
    testdir, plugin_options, get_vhost, rabbit_manager
):

    testdir.makepyfile(
        """
        import pytest

        from nameko.rpc import rpc
        from nameko.standalone.rpc import ServiceRpcClient

        class ServiceX(object):
            name = "x"

            @rpc
            def method(self):
                return "OK"

        @pytest.mark.usefixtures('rabbit_config')
        def test_runner(runner_factory):
            runner = runner_factory(ServiceX)
            runner.start()

            with ServiceRpcClient("x") as client:
                assert client.method() == "OK"
        """
    )
    result = testdir.runpytest(*plugin_options)
    assert result.ret == 0

    vhost = get_vhost(config['AMQP_URI'])
    assert get_rabbit_connections(vhost, rabbit_manager) == []


def test_runner_factory_with_config_patch_argument(testdir):

    testdir.makepyfile(
        """
        import nameko
        import pytest

        from nameko.testing.services import dummy

        class ServiceX(object):
            name = "x"

        class ServiceY(object):
            name = "y"

        def test_runner_factory_applies_config_patch_(runner_factory):

            assert "FOO" not in nameko.config

            runner = runner_factory({"FOO": 2}, ServiceX, ServiceY)
            runner.start()

            assert nameko.config["FOO"] == 2

        def test_runner_factory_rolls_back_config_patch():

            assert "FOO" not in nameko.config
        """
    )
    result = testdir.runpytest()
    assert result.ret == 0


@pytest.mark.usefixtures('rabbit_config')
@pytest.mark.usefixtures('predictable_call_ids')
def test_predictable_call_ids(runner_factory):

    worker_contexts = []

    class CaptureWorkerContext(DependencyProvider):
        def worker_setup(self, worker_ctx):
            worker_contexts.append(worker_ctx)

    class ServiceX(object):
        name = "x"

        capture = CaptureWorkerContext()
        service_y = ServiceRpc("y")

        @rpc
        def method(self):
            self.service_y.method()

    class ServiceY(object):
        name = "y"

        capture = CaptureWorkerContext()

        @rpc
        def method(self):
            pass

    runner = runner_factory(ServiceX, ServiceY)
    runner.start()

    with ServiceRpcClient("x") as service_x:
        service_x.method()

    call_ids = [worker_ctx.call_id for worker_ctx in worker_contexts]
    assert call_ids == ["x.method.1", "y.method.2"]


@pytest.mark.usefixtures('web_config')
def test_web_config():
    assert WEB_SERVER_CONFIG_KEY in config

    bind_address = parse_address(config[WEB_SERVER_CONFIG_KEY])
    sock = socket.socket()
    sock.bind(bind_address)


@pytest.mark.usefixtures('web_config')
def test_web_session(container_factory, web_session):

    class Service(object):
        name = "web"

        @http('GET', '/foo')
        def method(self, request):
            return "OK"

    container = container_factory(Service)
    container.start()

    assert web_session.get("/foo").status_code == 200


@pytest.mark.usefixtures('web_config')
def test_websocket(container_factory, websocket):

    class Service(object):
        name = "ws"

        @wsrpc
        def uppercase(self, socket_id, arg):
            return arg.upper()

    container = container_factory(Service)
    container.start()

    ws = websocket()
    assert ws.rpc("uppercase", arg="foo") == "FOO"



import logging
import time

import eventlet
import pytest
from mock import Mock

from nameko.testing.services import entrypoint_hook, get_extension
from nameko.testing.utils import wait_for_call
from nameko.timer import Timer, timer


@pytest.fixture
def tracker():
    return Mock()


@pytest.mark.parametrize('interval,eager,call_duration,expected_calls', [
    (0.1, False, 0, 1),
    (0.1, True, 0, 2),
    (5, False, 0, 0),
    (5, True, 0, 1),
    # call duration delays second call
    (0.1, False, 0.3, 1),  # fires at 0.1
    (0.1, True, 0.3, 1),  # fires at 0, 0.2
    (0.025, False, 0.07, 2),  # fires at 0.25, 0.095, 0.165
    (0.025, True, 0.07, 3),  # fires at 0, 0.07, 0.14
])
def test_timer_run(interval, eager, call_duration, expected_calls,
                   container_factory, tracker):
    """Test running the timers main loop.

    We test with "timer_only" mode, where only the main loop is run as well as
    in a more comprehensive mode where the entire container is tested.

    """
    timeout = 0.15
    times = []

    class Service(object):
        name = "service"

        @timer(interval, eager)
        def tick(self):
            times.append(time.time())
            eventlet.sleep(call_duration)
            tracker()

    container = container_factory(Service, {})

    # Check that Timer instance is initialized correctly
    instance = get_extension(container, Timer)
    assert instance.interval == interval
    assert instance.eager == eager

    t0 = time.time()
    container.start()
    eventlet.sleep(timeout)
    container.stop()

    rel_times = [t - t0 for t in times]
    assert tracker.call_count == expected_calls, (
        'Expected {} calls but got {} with '
        '{}timer interval of {} and call duration of {}. '
        'Times were {}'
    ).format(expected_calls, tracker.call_count,
             'eager-' if eager else '', interval, call_duration, rel_times)


def test_kill_stops_timer(container_factory, tracker):

    class Service(object):
        name = "service"

        @timer(0)
        def tick(self):
            tracker()

    container = container_factory(Service, {})
    container.start()

    with wait_for_call(1, tracker):
        container.kill()

    # unless the timer is dead, the following nap would cause a timer
    # to trigger
    eventlet.sleep(0.1)
    assert tracker.call_count == 1


def test_stop_while_sleeping(container_factory, tracker):
    """Check that waiting for the timer to fire does not block the container
    from being shut down gracefully.
    """
    class Service(object):
        name = "service"

        @timer(5)
        def tick(self):
            tracker()  # pragma: no cover

    container = container_factory(Service, {})
    container.start()

    # raise a Timeout if the container fails to stop within 1 second
    with eventlet.Timeout(1):
        container.stop()

    assert tracker.call_count == 0, 'Timer should not have fired.'


def test_timer_error(container_factory, caplog, tracker):
    """Check that an error in the decorated method does not cause the service
    containers loop to raise an exception.
    """

    class Service(object):
        name = "service"

        @timer(5, True)
        def tick(self):
            tracker()

    tracker.side_effect = ValueError('Boom!')
    container = container_factory(Service, {})

    with caplog.at_level(logging.CRITICAL):
        container.start()
        eventlet.sleep(0.05)
        # Check that the function was actually called and that the error was
        # handled gracefully.
        assert tracker.call_count == 1
        container.stop()

    # Check that no errors are thrown in the runners thread.
    # We can't check for raised errors here as the actual
    # exception is eaten by the worker pools handler.
    assert len(caplog.records) == 0, (
        'Expected no errors to have been '
        'raised in the worker thread.'
    )


def test_expected_error_in_worker(container_factory, caplog):
    """Make sure that expected exceptions are processed correctly."""
    class ExampleError(Exception):
        pass

    class Service(object):
        name = "service"

        @timer(1, expected_exceptions=(ExampleError,))
        def tick(self):
            raise ExampleError('boom!')

    container = container_factory(Service, {})

    with entrypoint_hook(container, "tick") as tick:
        container.start()
        with pytest.raises(ExampleError):
            tick()

    assert len(caplog.records) == 1
    assert caplog.records[0].message == (
        '(expected) error handling worker {}: boom!'
    ).format(caplog.records[0].args[0])


class TestEntrypointArguments:

    def test_expected_exceptions_and_sensitive_arguments(self, container_factory):

        class Boom(Exception):
            pass

        class Service(object):
            name = "service"

            @timer(1, expected_exceptions=Boom, sensitive_arguments=["arg"])
            def method(self, arg):
                pass  # pragma: no cover

        container = container_factory(Service)
        container.start()

        entrypoint = get_extension(container, Timer)
        assert entrypoint.expected_exceptions == Boom
        assert entrypoint.sensitive_arguments == ["arg"]



from collections import defaultdict

import pytest
from mock import call

from nameko.extensions import DependencyProvider, Extension, SharedExtension
from nameko.testing.utils import get_extension


class CallCollectorMixin(object):

    calls = defaultdict(lambda: defaultdict(list))

    def start(self, *args, **kwargs):
        self.calls[type(self)]['start'].append(call(*args, **kwargs))
        super(CallCollectorMixin, self).start(*args, **kwargs)


@pytest.yield_fixture(autouse=True)
def reset():
    yield
    CallCollectorMixin.calls.clear()


def test_simple_sharing(container_factory):

    class SimpleSharedExtension(CallCollectorMixin, SharedExtension):
        pass

    class SimpleDependencyProvider(CallCollectorMixin, DependencyProvider):
        ext = SimpleSharedExtension()

    class Service(object):
        name = "service"
        dep_1 = SimpleDependencyProvider()
        dep_2 = SimpleDependencyProvider()

    container = container_factory(Service)
    container.start()

    assert len(container.extensions) == 3
    calls = CallCollectorMixin.calls
    assert len(calls[SimpleDependencyProvider]['start']) == 2
    assert len(calls[SimpleSharedExtension]['start']) == 1


def test_custom_sharing_key(container_factory):

    class CustomSharedExtension(CallCollectorMixin, SharedExtension):
        def __init__(self, arg):
            self.arg = arg

        @property
        def sharing_key(self):
            return (type(self), self.arg)

    class SimpleDependencyProvider(CallCollectorMixin, DependencyProvider):
        ext_a = CustomSharedExtension("a")
        ext_b = CustomSharedExtension("b")

    class Service(object):
        name = "service"
        dep_1 = SimpleDependencyProvider()
        dep_2 = SimpleDependencyProvider()

    container = container_factory(Service)
    container.start()

    assert len(container.extensions) == 4
    calls = CallCollectorMixin.calls
    assert len(calls[SimpleDependencyProvider]['start']) == 2
    assert len(calls[CustomSharedExtension]['start']) == 2

    dep_1 = get_extension(
        container, SimpleDependencyProvider, attr_name="dep_1")
    dep_2 = get_extension(
        container, SimpleDependencyProvider, attr_name="dep_2")

    assert dep_1.ext_a is not dep_2.ext_b
    assert dep_1.ext_a is dep_2.ext_a
    assert dep_1.ext_b is dep_2.ext_b


def test_shared_intermediate(container_factory):

    class DedicatedExtension(CallCollectorMixin, Extension):
        pass

    class SharedIntermediate(CallCollectorMixin, SharedExtension):
        ext = DedicatedExtension()

    class SimpleDependencyProvider(CallCollectorMixin, DependencyProvider):
        ext = SharedIntermediate()

    class Service(object):
        name = "service"
        dep_1 = SimpleDependencyProvider()
        dep_2 = SimpleDependencyProvider()

    container = container_factory(Service)
    container.start()

    assert len(container.extensions) == 4
    calls = CallCollectorMixin.calls
    assert len(calls[SimpleDependencyProvider]['start']) == 2
    assert len(calls[SharedIntermediate]['start']) == 1
    assert len(calls[DedicatedExtension]['start']) == 1


def test_shared_extension_uniqueness(container_factory):

    class SimpleSharedExtension(CallCollectorMixin, SharedExtension):
        pass

    class SimpleDependencyProvider(CallCollectorMixin, DependencyProvider):
        ext = SimpleSharedExtension()

    class Service(object):
        name = "service"
        dep_1 = SimpleDependencyProvider()
        dep_2 = SimpleDependencyProvider()

    c1 = container_factory(Service)
    c2 = container_factory(Service)

    # extension declarations are identical between containers
    assert c1.service_cls.dep_1.ext == c2.service_cls.dep_1.ext

    # extension instances are different between containers
    shared_1 = get_extension(c1, SimpleSharedExtension)
    shared_2 = get_extension(c2, SimpleSharedExtension)
    assert shared_1 is not shared_2



from __future__ import absolute_import, unicode_literals

import sys
import time
import uuid
from functools import partial
from logging import getLogger

import kombu.serialization
from amqp.exceptions import NotFound
from eventlet.event import Event
from kombu import Exchange, Queue

from nameko import config, serialization
from nameko.amqp.consume import Consumer
from nameko.amqp.publish import Publisher, UndeliverableMessage, get_connection
from nameko.constants import (
    AMQP_SSL_CONFIG_KEY, AMQP_URI_CONFIG_KEY, DEFAULT_AMQP_URI,
    DEFAULT_HEARTBEAT, DEFAULT_PREFETCH_COUNT, HEARTBEAT_CONFIG_KEY,
    PREFETCH_COUNT_CONFIG_KEY, RPC_EXCHANGE_CONFIG_KEY
)
from nameko.exceptions import (
    ContainerBeingKilled, MalformedRequest, MethodNotFound,
    ReplyQueueExpiredWithPendingReplies, UnknownService,
    UnserializableValueError, deserialize, serialize
)
from nameko.extensions import (
    DependencyProvider, Entrypoint, ProviderCollector, SharedExtension
)
from nameko.messaging import decode_from_headers, encode_to_headers


_log = getLogger(__name__)


RPC_QUEUE_TEMPLATE = 'rpc-{}'
RPC_REPLY_QUEUE_TEMPLATE = 'rpc.reply-{}-{}'
RPC_REPLY_QUEUE_TTL = 300000  # ms (5 mins)

RESTRICTED_PUBLISHER_OPTIONS = (
    'exchange', 'routing_key', 'mandatory', 'reply_to', 'correlation_id'
)
"""
Publisher options that cannot be overridden when configuring an RPC client
"""


def get_rpc_exchange():
    # TODO: refactor this ugliness
    exchange_name = config.get(RPC_EXCHANGE_CONFIG_KEY, 'nameko-rpc')
    exchange = Exchange(exchange_name, durable=True, type="topic")
    return exchange


class RpcConsumer(SharedExtension, ProviderCollector):

    consumer_cls = Consumer

    def __init__(self, **consumer_options):
        self.queue = None
        self.consumer_options = consumer_options
        super(RpcConsumer, self).__init__()

    @property
    def amqp_uri(self):
        return config.get(AMQP_URI_CONFIG_KEY, DEFAULT_AMQP_URI)

    def setup(self):

        service_name = self.container.service_name
        queue_name = RPC_QUEUE_TEMPLATE.format(service_name)
        routing_key = '{}.*'.format(service_name)

        exchange = get_rpc_exchange()

        self.queue = Queue(
            queue_name,
            exchange=exchange,
            routing_key=routing_key,
            durable=True
        )

        ssl = config.get(AMQP_SSL_CONFIG_KEY)

        heartbeat = self.consumer_options.pop(
            'heartbeat', config.get(HEARTBEAT_CONFIG_KEY, DEFAULT_HEARTBEAT)
        )
        prefetch_count = self.consumer_options.pop(
            'prefetch_count', config.get(
                PREFETCH_COUNT_CONFIG_KEY, DEFAULT_PREFETCH_COUNT
            )
        )
        accept = self.consumer_options.pop(
            'accept', serialization.setup().accept
        )

        queues = [self.queue]
        callbacks = [self.handle_message]

        self.consumer = self.consumer_cls(
            self.amqp_uri, ssl=ssl, queues=queues, callbacks=callbacks,
            heartbeat=heartbeat, prefetch_count=prefetch_count,
            accept=accept
        )

    def start(self):
        self.container.spawn_managed_thread(self.consumer.run)
        self.consumer.wait_until_consumer_ready()

    def stop(self):
        self.consumer.stop()
        # super stop? (waits for all other providers)

    def unregister_provider(self, provider):
        self.stop()
        super(RpcConsumer, self).unregister_provider(provider)

    def get_provider_for_method(self, routing_key):
        service_name = self.container.service_name

        for provider in self._providers:
            key = '{}.{}'.format(service_name, provider.method_name)
            if key == routing_key:
                return provider
        else:
            method_name = routing_key.split(".")[-1]
            raise MethodNotFound(method_name)

    def handle_message(self, body, message):
        routing_key = message.delivery_info['routing_key']
        try:
            provider = self.get_provider_for_method(routing_key)
            provider.handle_message(body, message)
        except Exception:
            exc_info = sys.exc_info()
            self.handle_result(message, None, exc_info)

    def handle_result(self, message, result, exc_info):

        exchange = get_rpc_exchange()
        ssl = config.get(AMQP_SSL_CONFIG_KEY)

        responder = Responder(self.amqp_uri, exchange, message, ssl=ssl)
        result, exc_info = responder.send_response(result, exc_info)

        self.consumer.ack_message(message)

        return result, exc_info


class Rpc(Entrypoint):
    """
    A limitation of using a shared queue for all RPC entrypoints is
    that we can't accept per-entrypoint consumer options. The best solution
    to this is to start using a queue per entrypoint, but this will require
    a consumer (and if using kombu, a connection) per entrypoint.

    For the time being consumer options are not supported in RPC entrypoints.
    """
    rpc_consumer = RpcConsumer()

    def setup(self):
        self.rpc_consumer.register_provider(self)

    def stop(self):
        self.rpc_consumer.unregister_provider(self)

    def handle_message(self, body, message):
        try:
            args = body['args']
            kwargs = body['kwargs']
        except KeyError:
            raise MalformedRequest('Message missing `args` or `kwargs`')

        self.check_signature(args, kwargs)

        context_data = decode_from_headers(message.headers)

        handle_result = partial(self.handle_result, message)

        def spawn_worker():
            try:
                self.container.spawn_worker(
                    self, args, kwargs,
                    context_data=context_data,
                    handle_result=handle_result
                )
            except ContainerBeingKilled:
                self.rpc_consumer.consumer.requeue_message(message)

        service_name = self.container.service_name
        method_name = self.method_name

        # we must spawn a thread here to prevent handle_message blocking
        # when the worker pool is exhausted; if this happens the AMQP consumer
        # is also blocked and fails to send heartbeats, eventually causing it
        # to be disconnected
        # TODO replace global worker pool limits with per-entrypoint limits,
        # then remove this waiter thread
        ident = u"{}.wait_for_worker_pool[{}.{}]".format(
            type(self).__name__, service_name, method_name
        )
        self.container.spawn_managed_thread(spawn_worker, identifier=ident)

    def handle_result(self, message, worker_ctx, result, exc_info):
        result, exc_info = self.rpc_consumer.handle_result(
            message, result, exc_info)
        return result, exc_info


rpc = Rpc.decorator


class Responder(object):
    """ Helper object for publishing replies to RPC calls.
    """

    publisher_cls = Publisher

    def __init__(self, amqp_uri, exchange, message, ssl=None):
        self.amqp_uri = amqp_uri
        self.message = message
        self.exchange = exchange
        self.ssl = ssl

    def send_response(self, result, exc_info):

        error = None
        if exc_info is not None:
            error = serialize(exc_info[1])

        # send response encoded the same way as was the request message
        content_type = self.message.properties['content_type']
        serializer = kombu.serialization.registry.type_to_name[content_type]

        # disaster avoidance serialization check: `result` must be
        # serializable, otherwise the container will commit suicide assuming
        # unrecoverable errors (and the message will be requeued for another
        # victim)

        try:
            kombu.serialization.dumps(result, serializer)
        except Exception:
            exc_info = sys.exc_info()
            # `error` below is guaranteed to serialize to json
            error = serialize(UnserializableValueError(result))
            result = None

        payload = {'result': result, 'error': error}

        routing_key = self.message.properties['reply_to']
        correlation_id = self.message.properties.get('correlation_id')

        publisher = self.publisher_cls(self.amqp_uri, ssl=self.ssl)

        publisher.publish(
            payload,
            serializer=serializer,
            exchange=self.exchange,
            routing_key=routing_key,
            correlation_id=correlation_id
        )

        return result, exc_info


class ReplyListener(SharedExtension):
    """ SharedExtension for consuming replies to RPC requests.

    Creates a queue and consumes from it in a managed thread. RPC requests
    should register their `correlation_id` with
    :meth:`~ReplyListener.register_for_reply` in order for the `ReplyListener`
    to capture the reply.
    """

    class ReplyConsumer(Consumer):
        """ Subclass Consumer to add disconnection check
        """

        def __init__(self, check_for_lost_replies, *args, **kwargs):
            self.check_for_lost_replies = check_for_lost_replies
            super(ReplyListener.ReplyConsumer, self).__init__(*args, **kwargs)

        def get_consumers(self, consumer_cls, channel):
            """
            Check for messages lost while the reply listener was disconnected
            from the broker.
            """
            self.check_for_lost_replies()

            return super(ReplyListener.ReplyConsumer, self).get_consumers(
                consumer_cls, channel
            )

    consumer_cls = ReplyConsumer

    def __init__(self, **consumer_options):
        self.queue = None
        self.pending = {}
        self.consumer_options = consumer_options
        super(ReplyListener, self).__init__()

    @property
    def amqp_uri(self):
        return config.get(AMQP_URI_CONFIG_KEY, DEFAULT_AMQP_URI)

    def setup(self):

        reply_queue_uuid = uuid.uuid4()
        service_name = self.container.service_name

        queue_name = RPC_REPLY_QUEUE_TEMPLATE.format(
            service_name, reply_queue_uuid)

        self.routing_key = str(reply_queue_uuid)

        exchange = get_rpc_exchange()

        self.queue = Queue(
            queue_name,
            exchange=exchange,
            routing_key=self.routing_key,
            queue_arguments={
                'x-expires': RPC_REPLY_QUEUE_TTL
            }
        )

        ssl = config.get(AMQP_SSL_CONFIG_KEY)

        heartbeat = self.consumer_options.pop(
            'heartbeat', config.get(HEARTBEAT_CONFIG_KEY, DEFAULT_HEARTBEAT)
        )
        prefetch_count = self.consumer_options.pop(
            'prefetch_count', config.get(
                PREFETCH_COUNT_CONFIG_KEY, DEFAULT_PREFETCH_COUNT
            )
        )
        accept = self.consumer_options.pop(
            'accept', serialization.setup().accept
        )

        queues = [self.queue]
        callbacks = [self.handle_message]

        self.consumer = self.consumer_cls(
            self.check_for_lost_replies, self.amqp_uri, ssl=ssl,
            queues=queues, callbacks=callbacks,
            heartbeat=heartbeat, prefetch_count=prefetch_count, accept=accept
        )

    def start(self):
        self.container.spawn_managed_thread(self.consumer.run)
        self.consumer.wait_until_consumer_ready()

    def stop(self):
        self.consumer.stop()

    def check_for_lost_replies(self):
        if self.pending:
            try:
                ssl = config.get(AMQP_SSL_CONFIG_KEY)
                with get_connection(self.amqp_uri, ssl) as conn:
                    self.queue.bind(conn).queue_declare(passive=True)
            except NotFound:
                raise ReplyQueueExpiredWithPendingReplies(
                    "Lost replies for correlation ids:\n{}".format(
                        "\n".join(self.pending.keys())
                    )
                )

    def register_for_reply(self, correlation_id):
        """ Register an RPC call with the given `correlation_id` for a reply.

        Returns a function that can be used to retrieve the reply, blocking
        until it has been received.
        """
        reply_event = Event()
        self.pending[correlation_id] = reply_event
        return reply_event.wait

    def handle_message(self, body, message):
        self.consumer.ack_message(message)

        correlation_id = message.properties.get('correlation_id')
        rpc_call = self.pending.pop(correlation_id, None)
        if rpc_call is not None:
            rpc_call.send(body)
        else:
            _log.debug("Unknown correlation id: %s", correlation_id)


class ClusterRpc(DependencyProvider):
    """ DependencyProvider for injecting an RPC client to a cluster of
    services into a service.

    :Parameters:
        **publisher_options
            Options to configure the :class:`~nameko.amqqp.publish.Publisher`
            that sends the message.
    """

    publisher_cls = Publisher

    reply_listener = ReplyListener()

    def __init__(self, **publisher_options):

        for option in RESTRICTED_PUBLISHER_OPTIONS:
            publisher_options.pop(option, None)
        self.publisher_options = publisher_options

        default_uri = config.get(AMQP_URI_CONFIG_KEY, DEFAULT_AMQP_URI)
        self.amqp_uri = self.publisher_options.get('uri', default_uri)

    def setup(self):

        # ReplyListener.setup() runs concurrently in another thread, so
        # we need to wait here until it's defined its queue
        while self.reply_listener.queue is None:
            time.sleep(.1)

        exchange = get_rpc_exchange()

        default_serializer = self.container.serializer
        serializer = self.publisher_options.pop(
            'serializer', default_serializer
        )

        default_ssl = config.get(AMQP_SSL_CONFIG_KEY)
        ssl = self.publisher_options.pop('ssl', default_ssl)

        self.publisher_options.pop('uri', None)

        self.publisher = self.publisher_cls(
            self.amqp_uri,
            ssl=ssl,
            exchange=exchange,
            serializer=serializer,
            declare=[self.reply_listener.queue],
            reply_to=self.reply_listener.queue.routing_key,
            **self.publisher_options
        )

    def get_dependency(self, worker_ctx):

        publish = self.publisher.publish
        register_for_reply = self.reply_listener.register_for_reply

        return Client(publish, register_for_reply, worker_ctx.context_data)


class ServiceRpc(ClusterRpc):
    """ DependencyProvider for injecting an RPC client to a specific service
    into a service.

    As per :class:`~nameko.rpc.ClusterRpc` but with a pre-specified target
    service.

    :Parameters:
        target_service : str
            Target service name
    """

    def __init__(self, target_service, **kwargs):
        self.target_service = target_service
        super(ServiceRpc, self).__init__(**kwargs)

    def get_dependency(self, worker_ctx):
        client = super(ServiceRpc, self).get_dependency(worker_ctx)
        return getattr(client, self.target_service)


class Client(object):
    """ Helper object for making RPC calls.

    The target service and method name may be specified at construction time
    or by attribute or dict access, for example:

        # target at construction time
        client = Client(
            publish, register_for_reply, context_data,
            "target_service", "target_method"
        )
        client(*args, **kwargs)

        # equivalent with attribute access
        client = Client(publish, register_for_reply, context_data)
        client = client.target_service.target_method  # now fully-specified
        client(*args, **kwargs)

    Calling a fully-specified `Client` will make the RPC call and block for the
    result. The `call_async` method initiates the call but returns an
    `RpcReply` object that can be used later to retrieve the result.

    :Parameters:
        publish : callable
            Function to publish an RPC request
        register_for_reply : callable
            Function to register a new call with a reply listener. Returns
            another function that should be used to retrieve the response.
        context_data: dict
            Worker context data to be sent as extra headers
        service_name : str
            Optional target service name, if known
        method_name : str
            Optional target method name, if known
    """

    def __init__(
        self, publish, register_for_reply, context_data,
        service_name=None, method_name=None
    ):
        self.publish = publish
        self.register_for_reply = register_for_reply
        self.context_data = context_data
        self.service_name = service_name
        self.method_name = method_name

    def __getattr__(self, name):
        if self.method_name is not None:
            raise AttributeError(name)

        if self.service_name:
            target_service = self.service_name
            target_method = name
        else:
            target_service = name
            target_method = None

        clone = Client(
            self.publish,
            self.register_for_reply,
            self.context_data,
            target_service,
            target_method
        )
        return clone

    @property
    def fully_specified(self):
        return (
            self.service_name is not None and self.method_name is not None
        )

    @property
    def identifier(self):
        return "{}.{}".format(
            self.service_name or "*",
            self.method_name or "*"
        )

    def __getitem__(self, name):
        """Enable dict-like access on the client. """
        return getattr(self, name)

    def __call__(self, *args, **kwargs):
        rpc_call = self._call(*args, **kwargs)
        return rpc_call.result()

    def call_async(self, *args, **kwargs):
        rpc_call = self._call(*args, **kwargs)
        return rpc_call

    def _call(self, *args, **kwargs):
        if not self.fully_specified:
            raise ValueError(
                "Cannot call unspecified method {}".format(self.identifier)
            )

        _log.debug('invoking %s', self)

        # We use the `mandatory` flag in `producer.publish` below to catch rpc
        # calls to non-existent services, which would otherwise wait forever
        # for a reply that will never arrive.
        #
        # However, the basic.return ("no one is listening for topic") is sent
        # asynchronously and conditionally, so we can't wait() on the channel
        # for it (will wait forever on successful delivery).
        #
        # Instead, we make use of (the rabbitmq extension) confirm_publish
        # (https://www.rabbitmq.com/confirms.html), which _always_ sends a
        # reply down the channel. Moreover, in the case where no queues are
        # bound to the exchange (service unknown), the basic.return is sent
        # first, so by the time kombu returns (after waiting for the confim)
        # we can reliably check for returned messages.

        # Note that deactivating publish-confirms in the Client will disable
        # this functionality and therefore :class:`UnknownService` will never
        # be raised (and the caller will hang).

        correlation_id = str(uuid.uuid4())

        extra_headers = encode_to_headers(self.context_data)

        send_request = partial(
            self.publish, routing_key=self.identifier, mandatory=True,
            correlation_id=correlation_id, extra_headers=extra_headers
        )
        get_response = self.register_for_reply(correlation_id)

        rpc_call = RpcCall(correlation_id, send_request, get_response)

        try:
            rpc_call.send_request(*args, **kwargs)
        except UndeliverableMessage:
            raise UnknownService(self.service_name)

        return rpc_call


class RpcCall(object):
    """ Encapsulates a single RPC request and response.

    :Parameters:
        correlation_id : str
            Identifier for this call
        send_request : callable
            Function that initiates the request
        get_response : callable
            Function that retrieves the response
    """
    _response = None

    def __init__(self, correlation_id, send_request, get_response):
        self.correlation_id = correlation_id
        self._send_request = send_request
        self._get_response = get_response

    def send_request(self, *args, **kwargs):
        """ Send the RPC request to the remote service
        """
        payload = {'args': args, 'kwargs': kwargs}
        self._send_request(payload)

    def get_response(self):
        """ Retrieve the response for this RPC call. Blocks if the response
        has not been received.
        """
        if self._response is not None:
            return self._response

        self._response = self._get_response()
        return self._response

    def result(self):
        """ Return the result of this RPC call, blocking if the response
        has not been received.

        Raises a `RemoteError` if the remote service returned an error
        response.
        """
        response = self.get_response()
        error = response.get('error')
        if error:
            raise deserialize(error)
        return response['result']


RpcProxy = ServiceRpc  # backwards compat
Proxy = Client  # backwards compat



"""
Utilities for testing nameko services.
"""

import inspect
from collections import OrderedDict
from contextlib import contextmanager

import eventlet
from eventlet import event
from mock import MagicMock

from nameko.exceptions import ExtensionNotFound
from nameko.extensions import DependencyProvider, Entrypoint
from nameko.testing.utils import get_extension
from nameko.testing.waiting import WaitResult, wait_for_call


@contextmanager
def entrypoint_hook(container, method_name, context_data=None, timeout=30):
    """ Yield a function providing an entrypoint into a hosted service.

    The yielded function may be called as if it were the bare method defined
    in the service class. Intended to be used as an integration testing
    utility.

    :Parameters:
        container : ServiceContainer
            The container hosting the service owning the entrypoint
        method_name : str
            The name of the entrypoint decorated method on the service class
        context_data : dict
            Context data to provide for the call, e.g. a language, auth
            token or session.
        timeout : int
            Maximum seconds to wait

    **Usage**

    To verify that `ServiceX` and `ServiceY` are compatible, make an
    integration test that checks their interaction:

    .. literalinclude:: ../examples/testing/integration_x_y_test.py

    """
    entrypoint = get_extension(container, Entrypoint, method_name=method_name)
    if entrypoint is None:
        raise ExtensionNotFound(
            "No entrypoint for '{}' found on container {}.".format(
                method_name, container))

    def hook(*args, **kwargs):
        hook_result = event.Event()

        def wait_for_entrypoint():
            try:
                with entrypoint_waiter(
                    container, method_name,
                    timeout=timeout
                ) as waiter_result:
                    container.spawn_worker(
                        entrypoint, args, kwargs,
                        context_data=context_data
                    )
                hook_result.send(waiter_result.get())
            except Exception as exc:
                hook_result.send_exception(exc)

        def wait_for_container():
            try:
                container.wait()
            except Exception as exc:
                if not hook_result.ready():
                    hook_result.send_exception(exc)

        # If the container errors (e.g. due to a bad entrypoint), the
        # entrypoint_waiter never completes. To mitigate, we also wait on
        # the container, and if that throws we send the exception back
        # as our result.
        eventlet.spawn_n(wait_for_entrypoint)
        eventlet.spawn_n(wait_for_container)

        return hook_result.wait()

    yield hook


@contextmanager
def entrypoint_waiter(container, method_name, timeout=30, callback=None):
    """ Context manager that waits until an entrypoint has fired, and
    the generated worker has exited and been torn down.

    It yields a :class:`nameko.testing.waiting.WaitResult` object that can be
    used to get the result returned (exception raised) by the entrypoint
    after the waiter has exited.

    :Parameters:
        container : ServiceContainer
            The container hosting the service owning the entrypoint
        method_name : str
            The name of the entrypoint decorated method on the service class
        timeout : int
            Maximum seconds to wait
        callback : callable
            Function to conditionally control whether the entrypoint_waiter
            should exit for a particular invocation

    The `timeout` argument specifies the maximum number of seconds the
    `entrypoint_waiter` should wait before exiting. It can be disabled by
    passing `None`. The default is 30 seconds.

    Optionally allows a `callback` to be provided which is invoked whenever
    the entrypoint fires. If provided, the callback must return `True`
    for the `entrypoint_waiter` to exit. The signature for the callback
    function is::

        def callback(worker_ctx, result, exc_info):
            pass

    Where there parameters are as follows:

        `worker_ctx` (WorkerContext): WorkerContext of the entrypoint call.

        `result` (object): The return value of the entrypoint.

        `exc_info` (tuple): Tuple as returned by `sys.exc_info` if the
            entrypoint raised an exception, otherwise `None`.

    **Usage**

    ::

        class Service(object):
            name = "service"

            @event_handler('srcservice', 'eventtype')
            def handle_event(self, msg):
                return msg

        container = ServiceContainer(Service)
        container.start()

        # basic
        with entrypoint_waiter(container, 'handle_event'):
            ...  # action that dispatches event

        # giving access to the result
        with entrypoint_waiter(container, 'handle_event') as result:
            ...  # action that dispatches event
        res = result.get()

        # with custom timeout
        with entrypoint_waiter(container, 'handle_event', timeout=5):
            ...  # action that dispatches event

        # with callback that waits until entrypoint stops raising
        def callback(worker_ctx, result, exc_info):
            if exc_info is None:
                return True

        with entrypoint_waiter(container, 'handle_event', callback=callback):
            ...  # action that dispatches event

    """
    if not get_extension(container, Entrypoint, method_name=method_name):
        raise RuntimeError("{} has no entrypoint `{}`".format(
            container.service_name, method_name))

    class Result(WaitResult):
        worker_ctx = None

        def send(self, worker_ctx, result, exc_info):
            self.worker_ctx = worker_ctx
            super(Result, self).send(result, exc_info)

    waiter_callback = callback
    waiter_result = Result()

    def on_worker_result(worker_ctx, result, exc_info):
        complete = False
        if worker_ctx.entrypoint.method_name == method_name:
            if not callable(waiter_callback):
                complete = True
            else:
                complete = waiter_callback(worker_ctx, result, exc_info)

        if complete:
            waiter_result.send(worker_ctx, result, exc_info)
        return complete

    def on_worker_teardown(worker_ctx):
        if waiter_result.worker_ctx is worker_ctx:
            return True
        return False

    exc = entrypoint_waiter.Timeout(
        "Timeout on {}.{} after {} seconds".format(
            container.service_name, method_name, timeout)
    )

    with eventlet.Timeout(timeout, exception=exc):
        with wait_for_call(
            container, '_worker_teardown',
            lambda args, kwargs, res, exc: on_worker_teardown(*args)
        ):
            with wait_for_call(
                container, '_worker_result',
                lambda args, kwargs, res, exc: on_worker_result(*args)
            ):
                yield waiter_result


class EntrypointWaiterTimeout(Exception):
    pass


entrypoint_waiter.Timeout = EntrypointWaiterTimeout


def worker_factory(service_cls, **dependencies):
    """ Return an instance of ``service_cls`` with its injected dependencies
    replaced with :class:`~mock.MagicMock` objects, or as given in
    ``dependencies``.

    **Usage**

    The following example service proxies calls to a "maths" service via
    an ``ServiceRpc`` dependency::

        from nameko.rpc import ServiceRpc, rpc

        class ConversionService(object):
            name = "conversions"

            maths_rpc = ServiceRpc("maths")

            @rpc
            def inches_to_cm(self, inches):
                return self.maths_rpc.multiply(inches, 2.54)

            @rpc
            def cm_to_inches(self, cms):
                return self.maths_rpc.divide(cms, 2.54)

    Use the ``worker_factory`` to create an instance of
    ``ConversionService`` with its dependencies replaced by MagicMock objects::

        service = worker_factory(ConversionService)

    Nameko's entrypoints do not modify the service methods, so instance methods
    can be called directly with the same signature. The replaced dependencies
    can be used as any other MagicMock object, so a complete unit test for
    the conversion service may look like this::

        # create worker instance
        service = worker_factory(ConversionService)

        # replace "maths" service
        service.maths_rpc.multiply.side_effect = lambda x, y: x * y
        service.maths_rpc.divide.side_effect = lambda x, y: x / y

        # test inches_to_cm business logic
        assert service.inches_to_cm(300) == 762
        service.maths_rpc.multiply.assert_called_once_with(300, 2.54)

        # test cms_to_inches business logic
        assert service.cms_to_inches(762) == 300
        service.maths_rpc.divide.assert_called_once_with(762, 2.54)

    *Providing Dependencies*

    The ``**dependencies`` kwargs to ``worker_factory`` can be used to provide
    a replacement dependency instead of a mock. For example, to unit test a
    service against a real database:

    .. literalinclude::
        ../examples/testing/alternative_dependency_unit_test.py

    If a named dependency provider does not exist on ``service_cls``, a
    ``ExtensionNotFound`` exception is raised.

    """
    service = service_cls()
    for name, attr in inspect.getmembers(service_cls):
        if isinstance(attr, DependencyProvider):
            try:
                dependency = dependencies.pop(name)
            except KeyError:
                dependency = MagicMock()
            setattr(service, name, dependency)

    if dependencies:
        raise ExtensionNotFound(
            "DependencyProvider(s) '{}' not found on {}.".format(
                dependencies.keys(), service_cls))

    return service


class MockDependencyProvider(DependencyProvider):
    def __init__(self, attr_name, dependency=None):
        self.attr_name = attr_name
        self.dependency = MagicMock() if dependency is None else dependency

    def get_dependency(self, worker_ctx):
        return self.dependency


def _replace_dependencies(container, **dependency_map):
    if container.started:
        raise RuntimeError('You must replace dependencies before the '
                           'container is started.')

    dependency_names = {dep.attr_name for dep in container.dependencies}

    missing = set(dependency_map) - dependency_names
    if missing:
        raise ExtensionNotFound("Dependency(s) '{}' not found on {}.".format(
            missing, container))

    existing_providers = {dep.attr_name: dep for dep in container.dependencies
                          if dep.attr_name in dependency_map}

    for name, replacement in dependency_map.items():
        existing_provider = existing_providers[name]
        replacement_provider = MockDependencyProvider(
            name, dependency=replacement)
        container.dependencies.remove(existing_provider)
        container.dependencies.add(replacement_provider)


def replace_dependencies(container, *dependencies, **dependency_map):
    """ Replace the dependency providers on ``container`` with
    instances of :class:`MockDependencyProvider`.

    Dependencies named in *dependencies will be replaced with a
    :class:`MockDependencyProvider`, which injects a MagicMock instead of the
    dependency.

    Alternatively, you may use keyword arguments to name a dependency and
    provide the replacement value that the `MockDependencyProvider` should
    inject.

    Return the :attr:`MockDependencyProvider.dependency` for every dependency
    specified in the (*dependencies) args so that calls to the replaced
    dependencies can be inspected. Return a single object if only one
    dependency was replaced, and a generator yielding the replacements in the
    same order as ``dependencies`` otherwise.
    Note that any replaced dependencies specified via kwargs `**dependency_map`
    will not be returned.

    Replacements are made on the container instance and have no effect on the
    service class. New container instances are therefore unaffected by
    replacements on previous instances.

    **Usage**

    ::

        from nameko.rpc import ServiceRpc, rpc
        from nameko.standalone.rpc import ServiceRpcClient

        class ConversionService(object):
            name = "conversions"

            maths_rpc = ServiceRpc("maths")

            @rpc
            def inches_to_cm(self, inches):
                return self.maths_rpc.multiply(inches, 2.54)

            @rpc
            def cm_to_inches(self, cms):
                return self.maths_rpc.divide(cms, 2.54)

        container = ServiceContainer(ConversionService)
        mock_maths_rpc = replace_dependencies(container, "maths_rpc")
        mock_maths_rpc.divide.return_value = 39.37

        container.start()

        with ServiceRpcClient('conversions') as client:
            client.cm_to_inches(100)

        # assert that the dependency was called as expected
        mock_maths_rpc.divide.assert_called_once_with(100, 2.54)


    Providing a specific replacement by keyword:

    ::

        class StubMaths(object):

            def divide(self, val1, val2):
                return val1 / val2

        replace_dependencies(container, maths_rpc=StubMaths())

        container.start()

        with ServiceRpcClient('conversions') as client:
            assert client.cm_to_inches(127) == 50.0

    """
    if set(dependencies).intersection(dependency_map):
        raise RuntimeError(
            "Cannot replace the same dependency via both args and kwargs.")

    arg_replacements = OrderedDict((dep, MagicMock()) for dep in dependencies)

    dependency_map.update(arg_replacements)
    _replace_dependencies(container, **dependency_map)

    # if only one name was provided, return any replacement directly
    # otherwise return a generator
    res = (replacement for replacement in arg_replacements.values())
    if len(arg_replacements) == 1:
        return next(res)
    return res


def restrict_entrypoints(container, *entrypoints):
    """ Restrict the entrypoints on ``container`` to those named in
    ``entrypoints``.

    This method must be called before the container is started.

    **Usage**

    The following service definition has two entrypoints:

    .. code-block:: python

        class Service(object):
            name = "service"

            @timer(interval=1)
            def foo(self, arg):
                pass

            @rpc
            def bar(self, arg)
                pass

            @rpc
            def baz(self, arg):
                pass

        container = ServiceContainer(Service)

    To disable the timer entrypoint on ``foo``, leaving just the RPC
    entrypoints:

    .. code-block:: python

        restrict_entrypoints(container, "bar", "baz")

    Note that it is not possible to identify multiple entrypoints on the same
    method individually.

    """
    if container.started:
        raise RuntimeError('You must restrict entrypoints before the '
                           'container is started.')

    entrypoint_deps = list(container.entrypoints)
    entrypoint_names = {ext.method_name for ext in entrypoint_deps}

    missing = set(entrypoints) - entrypoint_names
    if missing:
        raise ExtensionNotFound("Entrypoint(s) '{}' not found on {}.".format(
            missing, container))

    for entrypoint in entrypoint_deps:
        if entrypoint.method_name not in entrypoints:
            container.entrypoints.remove(entrypoint)


class Once(Entrypoint):
    """ Entrypoint that spawns a worker exactly once, as soon as
    the service container started.
    """

    def __init__(self, *args, **kwargs):
        expected_exceptions = kwargs.pop('expected_exceptions', ())
        sensitive_arguments = kwargs.pop('sensitive_arguments', ())

        self.args = args
        self.kwargs = kwargs
        super(Once, self).__init__(
            expected_exceptions=expected_exceptions,
            sensitive_arguments=sensitive_arguments
        )

    def start(self):
        self.container.spawn_worker(self, self.args, self.kwargs)


once = Once.decorator

# dummy entrypoint
dummy = Entrypoint.decorator



# advanced_http.py

from nameko.web.handlers import http
from werkzeug.wrappers import Response

class Service:
    name = "advanced_http_service"

    @http('GET', '/privileged')
    def forbidden(self, request):
        return 403, "Forbidden"

    @http('GET', '/headers')
    def redirect(self, request):
        return 201, {'Location': 'https://www.example.com/widget/1'}, ""

    @http('GET', '/custom')
    def custom(self, request):
        return Response("payload")



from nameko.rpc import rpc

from .auth import Auth, Unauthenticated


class Service:
    name = "service"

    auth = Auth()

    @rpc(sensitive_arguments="password", expected_exceptions=Unauthenticated)
    def login(self, username, password):
        # raises Unauthenticated if username/password do not match
        return self.auth.authenticate(username, password)



# coding: utf-8

import pytest
from eventlet import GreenPool, sleep
from eventlet.event import Event

import nameko.rpc
from nameko.containers import ServiceContainer
from nameko.extensions import DependencyProvider
from nameko.rpc import Rpc, rpc
from nameko.testing.services import dummy, entrypoint_hook, get_extension
from nameko.utils import (
    REDACTED, get_redacted_args, import_from_path, sanitize_url
)
from nameko.utils.concurrency import fail_fast_imap


def test_fail_fast_imap():
    # A failing call...
    failing_exception = Exception()

    def failing_call():
        raise failing_exception

    # ...and an eventually successful call.
    slow_call_returned = Event()

    def slow_call():
        sleep(5)
        slow_call_returned.send()  # pragma: no cover

    def identity_fn(fn):
        return fn()

    calls = [slow_call, failing_call]

    pool = GreenPool(2)

    # fail_fast_imap fails as soon as the exception is raised
    with pytest.raises(Exception) as raised_exc:
        list(fail_fast_imap(pool, identity_fn, calls))
    assert raised_exc.value == failing_exception

    # The slow call won't go past the sleep as it was killed
    assert not slow_call_returned.ready()
    assert pool.free() == 2


class TestGetRedactedArgs(object):

    @pytest.mark.usefixtures("rabbit_config")
    @pytest.mark.parametrize("sensitive_arguments, expected", [
        (tuple(), {'a': 'A', 'b': 'B'}),  # no sensitive arguments
        ("a", {'a': REDACTED, 'b': 'B'}),
        (("a",), {'a': REDACTED, 'b': 'B'}),
        (("a", "b"), {'a': REDACTED, 'b': REDACTED}),
        (("c"), {'a': 'A', 'b': 'B'}),  # 'c' not a valid argument; ignored
    ])
    def test_get_redacted_args(self, sensitive_arguments, expected):

        class Service(object):
            name = "service"

            @rpc(sensitive_arguments=sensitive_arguments)
            def method(self, a, b):
                pass  # pragma: no cover

        args = ("A", "B")
        kwargs = {}

        container = ServiceContainer(Service)
        entrypoint = get_extension(container, Rpc)

        redacted = get_redacted_args(entrypoint, *args, **kwargs)
        assert redacted == expected

    @pytest.mark.usefixtures("rabbit_config")
    @pytest.mark.parametrize("args, kwargs", [
        (('A', "B"), {}),  # all args
        (('A',), {'b': "B"}),
        (tuple(), {'a': "A", 'b': "B"}),  # all kwargs
    ])
    def test_get_redacted_args_invocation(self, args, kwargs):

        class Service(object):
            name = "service"

            @rpc(sensitive_arguments="a")
            def method(self, a, b=None):
                pass  # pragma: no cover

        expected = {'a': REDACTED, 'b': 'B'}

        container = ServiceContainer(Service)
        entrypoint = get_extension(container, Rpc)

        redacted = get_redacted_args(entrypoint, *args, **kwargs)
        assert redacted == expected

    @pytest.mark.usefixtures("rabbit_config")
    @pytest.mark.parametrize("sensitive_arguments, expected", [
        (
            ("b.foo",),  # dict key
            {
                'a': "A",
                'b': {'foo': REDACTED, 'bar': "BAR"},
            },
        ),
        (
            ("b.foo[0]",),   # list index
            {
                'a': "A",
                'b': {'foo': [REDACTED, 2, 3], 'bar': "BAR"},
            }
        ),
        (
            ("b.not_a_key[0]",),  # missing keys ignored
            {
                'a': "A",
                'b': {'foo': [1, 2, 3], 'bar': "BAR"},
            }
        ),
        (
            ("b.foo[999]",),   # missing indices ignored
            {
                'a': "A",
                'b': {'foo': [1, 2, 3], 'bar': "BAR"},
            }
        ),
        (
            ("a.not_a_dictionary",),   # incorrect partials ignored
            {
                'a': "A",
                'b': {'foo': [1, 2, 3], 'bar': "BAR"},
            }
        ),
        (
            ("a", "b.foo[0]", "b.foo[2]", "b.bar"),  # multiple keys
            {
                'a': REDACTED,
                'b': {'foo': [REDACTED, 2, REDACTED], 'bar': REDACTED},
            }
        ),
    ])
    def test_get_redacted_args_partial(self, sensitive_arguments, expected):

        class Service(object):
            name = "service"

            @rpc(sensitive_arguments=sensitive_arguments)
            def method(self, a, b):
                pass  # pragma: no cover

        complex_arg = {
            'foo': [1, 2, 3],
            'bar': "BAR"
        }

        args = ("A", complex_arg)
        kwargs = {}

        container = ServiceContainer(Service)
        entrypoint = get_extension(container, Rpc)

        redacted = get_redacted_args(entrypoint, *args, **kwargs)
        assert redacted == expected

    def test_get_redacted_args_partial_no_mutation(self, container_factory):
        """ Calling `get_redacted_args` should not mutate anything received by
        the entrypoint.
        """
        redacted = {}

        class Redactor(DependencyProvider):
            """ Example DependencyProvider that redacts `sensitive_arguments`
            on entrypoints during the worker lifecycle.
            """

            def worker_setup(self, worker_ctx):
                entrypoint = worker_ctx.entrypoint
                args = worker_ctx.args
                kwargs = worker_ctx.kwargs

                redacted.update(get_redacted_args(entrypoint, *args, **kwargs))

        class Service(object):
            name = "service"

            redactor = Redactor()

            @dummy(sensitive_arguments=['b[foo][1]'])
            def method(self, a, b):
                return {
                    "a": a,
                    "b": b
                }

        container = container_factory(Service)

        complex_arg = {
            'foo': [1, 2, 3],
            'bar': "BAR"
        }

        with entrypoint_hook(container, 'method') as hook:
            result = hook("A", complex_arg)

        assert redacted == {
            'a': 'A',
            'b': {
                'foo': [1, REDACTED, 3],
                'bar': "BAR"
            }
        }
        assert result == {
            'a': 'A',
            'b': {
                'foo': [1, 2, 3],
                'bar': "BAR"
            }
        }


class TestImportFromPath(object):

    def test_path_is_none(self):
        assert import_from_path(None) is None

    def test_import_error(self):
        with pytest.raises(ImportError) as exc_info:
            import_from_path("foo.bar.Baz")
        assert (
            "`foo.bar.Baz` could not be imported" in str(exc_info.value)
        )

    def test_import_class(self):
        assert import_from_path("nameko.rpc.Rpc") is Rpc

    def test_import_module(self):
        assert import_from_path("nameko.rpc") is nameko.rpc

    def test_import_function(self):
        assert import_from_path("nameko.rpc.rpc") is rpc


@pytest.mark.parametrize('url,expected', [
    (
        'amqp://user:supersecret@127.0.0.1:5672//',
        'amqp://user:{}@127.0.0.1:5672//'.format(REDACTED)
    ),
    (
        'amqp://user@127.0.0.1:5672//',
        'amqp://user@127.0.0.1:5672//'
    ),
    (
        'amqp://127.0.0.1:5672//',
        'amqp://127.0.0.1:5672//'
    ),
])
def test_sanitize_url(url, expected):
    actual = sanitize_url(url)
    assert actual == expected



with ClusterRpcClient() as cluster_rpc:
    hello_res = cluster_rpc.service_x.remote_method.call_async("hello")
    world_res = cluster_rpc.service_x.remote_method.call_async("world")
    # do work while waiting
    hello_res.result()  # "hello-x-y"
    world_res.result()  # "world-x-y"



from copy import deepcopy

from six import wraps
from six.moves import UserDict


class Config(UserDict):

    def setup(self, config):
        self.data.clear()
        self.data.update(config)

    def patch(self, context_config, clear=False):
        """
        A context manager and decorator for updating config just for the given context
        or decorated function execution.

        Intended to be used mainly in tests:

            @nameko.config.patch({"AMQP_URI": "memory://"})
            def test_spam():
                assert nameko.config["AMQP_URI"] == "memory://"

        If `clear` is set to True then the config is completely replaced with the given
        `context_config` otherwise it's only updated.

        """

        data = self.data

        class Patcher:

            def __call__(self, func):
                @wraps(func)
                def wrapper(*args, **kwargs):
                    self.start()
                    try:
                        result = func(*args, **kwargs)
                    finally:
                        self.stop()
                    return result
                return wrapper

            def __init__(self, context_data, clear=False):
                self.clear = clear
                self.context_data = context_data

            def __enter__(self):
                self.start()

            def __exit__(self, *exc_info):
                self.stop()

            def start(self):
                self.original_data = deepcopy(data)
                if self.clear:
                    data.clear()
                data.update(self.context_data)

            def stop(self):
                data.clear()
                data.update(self.original_data)

        return Patcher(context_config, clear=clear)


config = Config()
"""
A dictionary holding and managing configuration

It is possible to set the config by using ``--config`` switch of Nameko
command line interface.

The config can be used straight on service definition level, e.g.::

    from nameko import config
    from nameko.messaging import Consumer, Queue

    class Service:

        @consume(
            queue=Queue(
                exchange=config.get("MY_EXCHANGE"),
                routing_key=config.get("MY_ROUTING_KEY"),
                name=config.get("MY_QUEUE_NAME")
            ),
            prefetch_count=config.get("MY_CONSUMER_PREFETCH_COUNT")
        )
        def consume(self, payload):
            pass

As a plain dictionary it is normally mutable at any point. After all,
we're all consenting adults here.)
"""



import bcrypt
import jwt

from nameko.extensions import DependencyProvider


JWT_SECRET = "secret"


class Unauthenticated(Exception):
    pass


class Unauthorized(Exception):
    pass


class Auth(DependencyProvider):

    class Api:
        def __init__(self, users, worker_ctx):
            self.users = users
            self.worker_ctx = worker_ctx

        def authenticate(self, username, password):
            user = self.users.get(username)
            if not user:
                raise Unauthenticated("User does not exist")
            if not bcrypt.checkpw(password.encode('utf-8'), user['password']):
                raise Unauthenticated("Incorrect password")

            payload = {
                'username': username,
                'roles': self.users[username]['roles']
            }
            token = jwt.encode(payload, key=JWT_SECRET).decode('utf-8')
            self.worker_ctx.context_data['auth'] = token
            return token

        def has_role(self, role):
            token = self.worker_ctx.context_data.get('auth')
            if not token:
                raise Unauthenticated()

            try:
                payload = jwt.decode(token, key=JWT_SECRET, verify=True)
                if role in payload['roles']:
                    return True
            except Exception:
                pass

            return False

        def check_role(self, role):
            if self.has_role(role):
                return
            raise Unauthorized()

    def setup(self):
        self.db = {
            'matt': {
                'password': (
                    b'$2b$12$fZXR7Z1Eoyn0pfym8.'
                    b'LyRuIFabYj00ZzhdaJ0qoTLZs9w4fg3pKlK'
                ),
                'roles': [
                    'developer',
                ]
            },
            'susie': {
                'password': (
                    b'$2b$12$k4MVi9PcbSsOqONoj5vW9.'
                    b'pcQpB0xSjYkZcc6Ogr5nQ4MD8DRDiUK'
                ),
                'roles': [
                    'developer',
                    'admin'
                ]
            }
        }

    def get_dependency(self, worker_ctx):
        return Auth.Api(self.db, worker_ctx)



import logging

import pytest
from mock import ANY, Mock, call, patch

from nameko.log_helpers import make_timing_logger


@pytest.yield_fixture
def mock_logging():
    with patch('nameko.log_helpers.logging') as patched:
        yield patched


@pytest.yield_fixture
def mock_time():
    with patch('nameko.log_helpers.time') as patched:
        patched.time.side_effect = [0, 0.123456789]
        yield


def test_timing_logger(mock_logging, mock_time):
    logger = Mock()
    log_time = make_timing_logger(logger)

    with log_time("msg %s", "foo"):
        pass

    assert logger.log.call_args_list == [
        call(logging.DEBUG, "msg %s in %0.3fs", "foo", 0.123456789,)
    ]


def test_timing_logger_level(mock_logging):

    logger = Mock()
    log_time = make_timing_logger(logger, level=logging.INFO)

    with log_time("msg %s", "foo"):
        pass

    assert logger.log.call_args_list == [
        call(logging.INFO, "msg %s in %0.3fs", "foo", ANY)
    ]


def test_timing_logger_precision(mock_logging):

    logger = Mock()
    log_time = make_timing_logger(logger, precision=5)

    with log_time("msg %s", "foo"):
        pass

    assert logger.log.call_args_list == [
        call(logging.DEBUG, "msg %s in %0.5fs", "foo", ANY)
    ]



"""
Common testing utilities.
"""
import socket
from contextlib import contextmanager
from functools import partial

import eventlet

from nameko.extensions import Entrypoint
from nameko.testing.rabbit import HTTPError


def get_extension(container, extension_cls, **match_attrs):
    """ Inspect ``container.extensions`` and return the first item that is
    an instance of ``extension_cls``.

    Optionally also require that the instance has an attribute with a
    particular value as given in the ``match_attrs`` kwargs.
    """
    for ext in container.extensions:
        if isinstance(ext, extension_cls):
            if not match_attrs:
                return ext

            def has_attribute(name, value):
                return getattr(ext, name) == value

            if all([has_attribute(name, value)
                    for name, value in match_attrs.items()]):
                return ext


def get_container(runner, service_cls):
    """ Inspect ``runner.containers`` and return the first item that is
    hosting an instance of ``service_cls``.
    """
    for container in runner.containers:
        if container.service_cls == service_cls:
            return container


@contextmanager
def wait_for_call(timeout, mock_method):
    """ Return a context manager that waits ``timeout`` seconds for
    ``mock_method`` to be called, yielding the mock if so.

    Raises an :class:`eventlet.Timeout` if the method was not called
    within ``timeout`` seconds.
    """
    with eventlet.Timeout(timeout):
        while not mock_method.called:
            eventlet.sleep()
    yield mock_method


def assert_stops_raising(fn, exception_type=Exception, timeout=10,
                         interval=0.1):
    """Assert that ``fn`` returns successfully within ``timeout``
       seconds, trying every ``interval`` seconds.

       If ``exception_type`` is provided, fail unless the exception thrown is
       an instance of ``exception_type``. If not specified, any
       `:class:`Exception` instance is allowed.
    """
    with eventlet.Timeout(timeout):
        while True:
            try:
                fn()
            except exception_type:
                pass
            else:
                return
            eventlet.sleep(interval)


class AnyInstanceOf(object):

    def __init__(self, cls):
        self.cls = cls

    def __eq__(self, other):
        return isinstance(self.cls, type) and isinstance(other, self.cls)

    def __ne__(self, other):
        return not (self == other)

    def __repr__(self):
        obj = getattr(self.cls, '__name__', self.cls)
        return '<AnyInstanceOf-{}>'.format(obj)


ANY_PARTIAL = AnyInstanceOf(partial)


class DummyProvider(Entrypoint):
    def __init__(self, method_name=None):
        self.method_name = method_name


def get_rabbit_connections(vhost, rabbit_manager):

    connections = rabbit_manager.get_connections()
    if connections is not None:
        return [connection for connection in connections
                if connection['vhost'] == vhost]
    return []


def reset_rabbit_connections(vhost, rabbit_manager):

    for connection in get_rabbit_connections(vhost, rabbit_manager):
        try:
            rabbit_manager.delete_connection(connection['name'])
        except HTTPError as exc:
            if exc.response.status_code == 404:
                pass  # connection closed in a race
            else:
                raise


def find_free_port(host='127.0.0.1'):
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock.bind((host, 0))
    port = sock.getsockname()[1]
    sock.close()
    return port


class ResourcePipeline(object):
    """ Creates and destroys resources in background threads.

    Creates up to `size` resources ahead of time so the caller avoids waiting
    for lazy creation.
    """

    STOP = object()

    def __init__(self, create, destroy, size=3):
        if size == 0:
            raise RuntimeError("Zero size would create unbounded resources")
        self.ready = eventlet.Queue(maxsize=size)
        self.trash = eventlet.Queue()
        self.size = size

        self.create = create
        self.destroy = destroy

    def _start(self):
        self.running = True
        self.create_thread = eventlet.spawn(self._create)
        self.destroy_thread = eventlet.spawn(self._destroy)

    def _shutdown(self):
        self.running = False

        # increase max size of the ready queue and yield, allowing the
        # create thread to exit now if it's blocked trying to put an item
        self.ready.resize(self.size + 1)
        eventlet.sleep()

        # trash unused items while there are any left in the queue,
        # or the create thread is still running
        while self.ready.qsize() or not self.create_thread.dead:
            unused = self.ready.get()
            self.trash.put(unused)

        # finally wait for the destroy thread to exit
        self.trash.put(ResourcePipeline.STOP)
        self.destroy_thread.wait()

    def _create(self):
        while self.running:
            item = self.create()
            self.ready.put(item)

    def _destroy(self):
        while True:
            item = self.trash.get()
            if item is ResourcePipeline.STOP:
                break
            self.destroy(item)

    @contextmanager
    def get(self):
        item = self.ready.get()
        try:
            yield item
        finally:
            self.trash.put(item)

    @contextmanager
    def run(self):
        self._start()
        try:
            yield self
        finally:
            self._shutdown()



import pytest
from mock import call, patch

from nameko import config
from nameko.dependency_providers import Config
from nameko.rpc import rpc
from nameko.testing.services import entrypoint_hook


class TestConfig:

    @pytest.fixture
    def service_cls(self):
        class Service:
            name = "test_x"
            config = Config()

            @rpc
            def foo(self):
                return self.config.get("FOO")

        return Service

    @pytest.mark.usefixtures("rabbit_config")
    @patch("nameko.dependency_providers.warnings.warn")
    def test_get_config_value(
        self, warn, rabbit_config, container_factory, service_cls
    ):
        with config.patch({"FOO": "bar"}):

            container = container_factory(service_cls)
            container.start()

            with entrypoint_hook(container, "foo") as foo:
                assert foo() == "bar"

        assert (
            warn.call_args ==
            call("Use ``nameko.config`` instead.", DeprecationWarning)
        )



import pytest

from nameko.contextdata import Language
from nameko.rpc import rpc
from nameko.testing.services import entrypoint_hook


class HelloService:
    """ Service under test
    """
    name = "hello_service"

    language = Language()

    @rpc
    def hello(self, name):
        greeting = "Hello"
        if self.language == "fr":
            greeting = "Bonjour"
        elif self.language == "de":
            greeting = "Gutentag"

        return "{}, {}!".format(greeting, name)


@pytest.mark.usefixtures("rabbit_config")
@pytest.mark.parametrize("language, greeting", [
    ("en", "Hello"),
    ("fr", "Bonjour"),
    ("de", "Gutentag"),
])
def test_hello_languages(language, greeting, container_factory):

    container = container_factory(HelloService)
    container.start()

    context_data = {'language': language}
    with entrypoint_hook(container, 'hello', context_data) as hook:
        assert hook("Matt") == "{}, Matt!".format(greeting)



# coding: utf-8

import sys
from functools import partial

import eventlet
import greenlet
import pytest
from eventlet import Timeout, sleep, spawn
from eventlet.event import Event
from mock import ANY, Mock, call, patch

from nameko import config
from nameko.constants import MAX_WORKERS_CONFIG_KEY
from nameko.containers import ServiceContainer, get_service_name
from nameko.exceptions import ConfigurationError
from nameko.extensions import DependencyProvider, Entrypoint
from nameko.testing.utils import get_extension


class CallCollectorMixin(object):
    call_counter = 0

    def __init__(self):
        self._reset_calls()

    def bind(self, *args):
        res = super(CallCollectorMixin, self).bind(*args)
        self.instances.add(res)
        return res

    def _reset_calls(self):
        self.calls = []
        self.call_ids = []

    def _log_call(self, data):
        CallCollectorMixin.call_counter += 1
        self.calls.append(data)
        self.call_ids.append(CallCollectorMixin.call_counter)

    def setup(self):
        self._log_call(('setup'))
        super(CallCollectorMixin, self).setup()

    def start(self):
        self._log_call(('start'))
        super(CallCollectorMixin, self).start()

    def stop(self):
        self._log_call(('stop'))
        super(CallCollectorMixin, self).stop()

    def kill(self, exc=None):
        self._log_call(('kill'))
        super(CallCollectorMixin, self).stop()


class CallCollectingEntrypoint(CallCollectorMixin, Entrypoint):
    instances = set()
    expected_exceptions = ()


class CallCollectingDependencyProvider(CallCollectorMixin, DependencyProvider):
    instances = set()

    def get_dependency(self, worker_ctx):
        self._log_call(('get_dependency', worker_ctx))
        return 'spam-attr'

    def worker_setup(self, worker_ctx):
        self._log_call(('worker_setup', worker_ctx))
        super(CallCollectorMixin, self).worker_setup(worker_ctx)

    def worker_result(self, worker_ctx, result=None, exc_info=None):
        self._log_call(('worker_result', worker_ctx, (result, exc_info)))
        super(CallCollectorMixin, self).worker_result(
            worker_ctx, result, exc_info)

    def worker_teardown(self, worker_ctx):
        self._log_call(('worker_teardown', worker_ctx))
        super(CallCollectorMixin, self).worker_teardown(worker_ctx)


foobar = CallCollectingEntrypoint.decorator

egg_error = Exception('broken')


class Service(object):
    name = 'test-service'

    spam = CallCollectingDependencyProvider()

    @foobar
    def ham(self):
        return 'ham'

    @foobar
    def egg(self):
        raise egg_error


@pytest.fixture
def container():
    container = ServiceContainer(Service)
    for ext in container.extensions:
        ext._reset_calls()

    CallCollectorMixin.call_counter = 0
    return container


@pytest.yield_fixture
def logger():
    with patch('nameko.containers._log', autospec=True) as patched:
        yield patched


def test_collects_extensions(container):
    assert len(container.extensions) == 3
    assert container.extensions == (
        CallCollectingEntrypoint.instances |
        CallCollectingDependencyProvider.instances)


def test_starts_extensions(container):

    for ext in container.extensions:
        assert ext.calls == []

    container.start()

    for ext in container.extensions:
        assert ext.calls == [
            ('setup'),
            ('start')
        ]


def test_stops_extensions(container):

    container.stop()
    for ext in container.extensions:
        assert ext.calls == [
            ('stop')
        ]


def test_stops_entrypoints_before_dependency_providers(container):
    container.stop()

    provider = get_extension(container, DependencyProvider)

    for entrypoint in container.entrypoints:
        assert entrypoint.call_ids[0] < provider.call_ids[0]


def test_worker_life_cycle(container):

    spam_dep = get_extension(container, DependencyProvider)
    ham_dep = get_extension(container, Entrypoint, method_name="ham")
    egg_dep = get_extension(container, Entrypoint, method_name="egg")

    handle_result = Mock()
    handle_result.side_effect = (
        lambda worker_ctx, res, exc_info: (res, exc_info))

    ham_worker_ctx = container.spawn_worker(
        ham_dep, [], {}, handle_result=handle_result)
    container._worker_pool.waitall()

    egg_worker_ctx = container.spawn_worker(
        egg_dep, [], {}, handle_result=handle_result)
    container._worker_pool.waitall()

    assert spam_dep.calls == [
        ('get_dependency', ham_worker_ctx),
        ('worker_setup', ham_worker_ctx),
        ('worker_result', ham_worker_ctx, ('ham', None)),
        ('worker_teardown', ham_worker_ctx),
        ('get_dependency', egg_worker_ctx),
        ('worker_setup', egg_worker_ctx),
        ('worker_result', egg_worker_ctx, (None, (Exception, egg_error, ANY))),
        ('worker_teardown', egg_worker_ctx),
    ]

    assert handle_result.call_args_list == [
        call(ham_worker_ctx, "ham", None),
        call(egg_worker_ctx, None, (Exception, egg_error, ANY))
    ]


def test_wait_waits_for_container_stopped(container):
    gt = spawn(container.wait)

    with Timeout(1):
        assert not gt.dead
        container.stop()
        sleep(0.01)
        assert gt.dead


@config.patch({MAX_WORKERS_CONFIG_KEY: 1}, clear=True)
def test_container_doesnt_exhaust_max_workers(container):
    spam_called = Event()
    spam_continue = Event()

    class Service(object):
        name = 'max-workers'

        @foobar
        def spam(self, a):
            spam_called.send(a)
            spam_continue.wait()

    container = ServiceContainer(Service)

    dep = get_extension(container, Entrypoint)

    # start the first worker, which should wait for spam_continue
    container.spawn_worker(dep, ['ham'], {})

    # start the next worker in a speparate thread,
    # because it should block until the first one completed
    gt = spawn(container.spawn_worker, dep, ['eggs'], {})

    with Timeout(1):
        assert spam_called.wait() == 'ham'
        # if the container had spawned the second worker, we would see
        # an error indicating that spam_called was fired twice, and the
        # greenthread would now be dead.
        assert not gt.dead
        # reset the calls and allow the waiting worker to complete.
        spam_called.reset()
        spam_continue.send(None)
        # the second worker should now run and complete
        assert spam_called.wait() == 'eggs'
        assert gt.dead


def test_stop_already_stopped(container, logger):

    assert not container._died.ready()
    container.stop()
    assert container._died.ready()

    container.stop()
    assert logger.debug.call_args == call("already stopped %s", container)


def test_kill_already_stopped(container, logger):

    assert not container._died.ready()
    container.stop()
    assert container._died.ready()

    container.kill()
    assert logger.debug.call_args == call("already stopped %s", container)


def test_kill_container_with_managed_threads(container):
    """ Start a thread that's not managed by dependencies. Ensure it is killed
    when the container is.
    """
    def sleep_forever():
        while True:
            sleep()

    container.spawn_managed_thread(sleep_forever)

    assert len(container._managed_threads) == 1
    (worker_gt,) = container._managed_threads.keys()

    container.kill()

    with Timeout(1):
        container._died.wait()

        with pytest.raises(greenlet.GreenletExit):
            worker_gt.wait()


def test_kill_container_with_active_workers(container_factory):
    waiting = Event()
    wait_forever = Event()

    class Service(object):
        name = 'kill-with-active-workers'

        @foobar
        def spam(self):
            waiting.send(None)
            wait_forever.wait()

    container = container_factory(Service)
    dep = get_extension(container, Entrypoint)

    # start the first worker, which should wait for spam_continue
    container.spawn_worker(dep, (), {})

    waiting.wait()

    with patch('nameko.containers._log') as logger:
        container.kill()

    assert logger.warning.call_args_list == [
        call('killing %s active workers(s)', 1),
        call('killing active worker for %s', ANY)
    ]


def test_handle_killed_worker(container, logger):

    dep = get_extension(container, Entrypoint)
    container.spawn_worker(dep, ['sleep'], {})

    assert len(container._worker_threads) == 1
    (worker_gt,) = container._worker_threads.values()

    worker_gt.kill()
    assert logger.debug.call_args == call(
        "%s thread killed by container", container)

    assert not container._died.ready()  # container continues running


def test_spawned_thread_kills_container(container):
    def raise_error():
        raise Exception('foobar')

    container.start()
    container.spawn_managed_thread(raise_error)

    with pytest.raises(Exception) as exc_info:
        container.wait()

    assert exc_info.value.args == ('foobar',)


def test_spawned_thread_causes_container_to_kill_other_thread(container):
    killed_by_error_raised = Event()

    def raise_error():
        raise Exception('foobar')

    def wait_forever():
        try:
            Event().wait()
        except:
            killed_by_error_raised.send()
            raise

    container.start()

    container.spawn_managed_thread(wait_forever)
    container.spawn_managed_thread(raise_error)

    with Timeout(1):
        killed_by_error_raised.wait()


def test_container_only_killed_once(container):

    class Broken(Exception):
        pass

    exc = Broken('foobar')

    def raise_error():
        raise exc

    with patch.object(
        container, '_kill_managed_threads', autospec=True
    ) as kill_managed_threads:

        with patch.object(container, 'kill', wraps=container.kill) as kill:

            container.start()

            # any exception in a managed thread will cause the container
            # to be killed. Two threads raising an exception will cause
            # the container's handle_thread_exit() method to kill the
            # container twice.
            container.spawn_managed_thread(raise_error)
            container.spawn_managed_thread(raise_error)

            # container should die with an exception
            with pytest.raises(Broken):
                container.wait()

            # container.kill should have been called twice
            assert kill.call_args_list == [
                call((Broken, exc, ANY)),
                call((Broken, exc, ANY))
            ]

            # only the first kill results in any cleanup
            assert kill_managed_threads.call_count == 1


def test_container_stop_kills_remaining_managed_threads(container, logger):
    """ Verify any remaining managed threads are killed when a container
    is stopped.
    """
    def sleep_forever():
        while True:
            sleep()

    container.start()

    container.spawn_managed_thread(sleep_forever)
    container.spawn_managed_thread(sleep_forever)

    container.stop()

    assert logger.warning.call_args_list == [
        call("killing %s managed thread(s)", 2),
        call("killing managed thread `%s`", "sleep_forever"),
        call("killing managed thread `%s`", "sleep_forever"),
    ]

    assert logger.debug.call_args_list == [
        call("starting %s", container),
        call("stopping %s", container),
        call("%s thread killed by container", container),
        call("%s thread killed by container", container),
    ]


def test_container_kill_kills_remaining_managed_threads(container, logger):
    """ Verify any remaining managed threads are killed when a container
    is killed.
    """
    def sleep_forever():
        while True:
            sleep()

    container.start()

    container.spawn_managed_thread(sleep_forever)
    container.spawn_managed_thread(sleep_forever)

    container.kill()

    assert logger.warning.call_args_list == [
        call("killing %s managed thread(s)", 2),
        call("killing managed thread `%s`", "sleep_forever"),
        call("killing managed thread `%s`", "sleep_forever"),
    ]

    assert logger.info.call_args_list == [
        call("killing %s", container),
    ]

    assert logger.debug.call_args_list == [
        call("starting %s", container),
        call("%s thread killed by container", container),
        call("%s thread killed by container", container),
    ]


def test_kill_bad_dependency(container):
    """ Verify that an exception from a badly-behaved dependency.kill()
    doesn't stop the container's kill process.
    """
    dep = get_extension(container, DependencyProvider)
    with patch.object(dep, 'kill') as dep_kill:
        dep_kill.side_effect = Exception('dependency error')

        container.start()

        # manufacture an exc_info to kill with
        try:
            raise Exception('container error')
        except:
            exc_info = sys.exc_info()

        container.kill(exc_info)

        with pytest.raises(Exception) as exc_info:
            container.wait()
        assert str(exc_info.value) == "container error"


def test_stop_during_kill(container, logger):
    """ Verify we handle the race condition when a runner tries to stop
    a container while it is being killed.
    """
    with patch.object(
        container, '_kill_managed_threads', autospec=True
    ) as kill_managed_threads:

        # force eventlet yield during kill() so stop() will be scheduled
        kill_managed_threads.side_effect = eventlet.sleep

        # manufacture an exc_info to kill with
        try:
            raise Exception('error')
        except:
            exc_info = sys.exc_info()

        eventlet.spawn(container.kill, exc_info)
        eventlet.spawn(container.stop)

        with pytest.raises(Exception):
            container.wait()
        assert logger.debug.call_args_list == [
            call("already being killed %s", container),
        ]


def test_get_service_name():

    class Service():
        name = "str"

    class UnicodeService():
        name = u"unicøde"

    class BadNameService():
        name = object()

    class AnonymousService():
        pass

    assert get_service_name(Service) == "str"
    assert get_service_name(UnicodeService) == u"unicøde"

    with pytest.raises(ConfigurationError) as exc_info:
        get_service_name(BadNameService)
    assert str(exc_info.value) == (
        'Service name attribute must be a string '
        '(test.test_container.BadNameService.name)'
    )

    with pytest.raises(ConfigurationError) as exc_info:
        get_service_name(AnonymousService)
    assert str(exc_info.value) == (
        'Service class must define a `name` attribute '
        '(test.test_container.AnonymousService)'
    )


def test_logging_managed_threads(container, logger):

    def wait():
        Event().wait()

    container.spawn_managed_thread(wait)  # anonymous
    container.spawn_managed_thread(partial(wait))  # __name__less
    container.spawn_managed_thread(wait, identifier='named')

    container.stop()

    # order is not guaranteed when killing threads
    call_args_list = logger.warning.call_args_list
    assert call("killing %s managed thread(s)", 3) in call_args_list
    assert call("killing managed thread `%s`", "wait") in call_args_list
    assert call("killing managed thread `%s`", "<unknown>") in call_args_list
    assert call("killing managed thread `%s`", "named") in call_args_list


@config.patch({'spam': 'ham'}, clear=True)
def test_container_config_deprecated(container):

    with patch('nameko.containers.warnings') as warnings:
        container.config == config

    assert warnings.warn.call_count



import json
import sys

import pytest
from mock import ANY

from nameko.exceptions import RemoteError
from nameko.extensions import DependencyProvider
from nameko.rpc import Rpc
from nameko.standalone.rpc import ServiceRpcClient
from nameko.testing.services import entrypoint_waiter


worker_result_called = []


@pytest.yield_fixture(autouse=True)
def reset():
    yield
    del worker_result_called[:]


class ResultCollector(DependencyProvider):
    """ DependencyProvider that collects worker results
    """
    def worker_result(self, worker_ctx, res, exc_info):
        worker_result_called.append((res, exc_info))


class CustomRpc(Rpc):
    """ Rpc subclass that verifies `result` can be serialized to json,
    and changes the `result` and `exc_info` accordingly.
    """
    def handle_result(self, message, worker_ctx, result, exc_info):
        try:
            json.dumps(result)
        except Exception:
            result = "something went wrong"
            exc_info = sys.exc_info()

        return super(CustomRpc, self).handle_result(
            message, worker_ctx, result, exc_info)


custom_rpc = CustomRpc.decorator


class ExampleService(object):
    name = "exampleservice"

    collector = ResultCollector()

    @custom_rpc
    def echo(self, arg):
        return arg

    @custom_rpc
    def unserializable(self):
        return object()


@pytest.yield_fixture
def rpc_client(rabbit_config):
    with ServiceRpcClient('exampleservice') as client:
        yield client


@pytest.mark.usefixtures("rabbit_config")
def test_handle_result(
    container_factory, rabbit_manager, rpc_client
):
    """ Verify that `handle_result` can modify the return values of the worker,
    such that other dependencies see the updated values.
    """
    container = container_factory(ExampleService)
    container.start()

    assert rpc_client.echo("hello") == "hello"

    # use entrypoint_waiter because the rpc client returns before
    # worker_result is called
    with entrypoint_waiter(container, "unserializable"):
        with pytest.raises(RemoteError) as exc:
            rpc_client.unserializable()
        assert "is not JSON serializable" in str(exc.value)

    # verify ResultCollector sees values returned from `handle_result`
    assert worker_result_called == [
        ("hello", None),
        ("something went wrong", (TypeError, ANY, ANY)),
    ]



"""
Provides a high level interface to the core messaging module.

Events are special messages, which can be emitted by one service
and handled by other listening services.

An event consists of an identifier and some data and is dispatched using an
injection acquired from an instance of :class:`EventDispatcher`.

Events are dispatched asynchronously. It is only guaranteed that the event has
been dispatched, not that it was received or handled by a listener.

To listen to an event, a service must declare a handler using the
:func:`handle_event` entrypoint, providing the target service and an event type
filter.

Example::

    # service A
    def edit_foo(self, id):
        # ...
        self.dispatch('foo_updated', {'id': id})

    # service B

    @handle_event('service_a', 'foo_updated')
    def bar(event_data):
        pass

"""
from __future__ import absolute_import

import uuid
from logging import getLogger

from kombu import Queue

from nameko.messaging import Consumer, Publisher, encode_to_headers
from nameko.standalone.events import get_event_exchange


SERVICE_POOL = "service_pool"
SINGLETON = "singleton"
BROADCAST = "broadcast"

_log = getLogger(__name__)


class EventHandlerConfigurationError(Exception):
    """ Raised when an event handler is misconfigured.
    """


class EventDispatcher(Publisher):
    """ Provides an event dispatcher method via dependency injection.

    Events emitted will be dispatched via the service's events exchange,
    which automatically gets declared by the event dispatcher
    as a topic exchange.
    The name for the exchange will be `{service-name}.events`.

    Events, emitted via the dispatcher, will be serialized and published
    to the events exchange. The event's type attribute is used as the
    routing key, which can be used for filtering on the listener's side.

    The dispatcher will return as soon as the event message has been published.
    There is no guarantee that any service will receive the event, only
    that the event has been successfully dispatched.

    Example::

        class Spammer(object):
            dispatch_spam = EventDispatcher()

            def emit_spam(self):
                evt_data = 'ham and eggs'
                self.dispatch_spam('spam.ham', evt_data)

    """

    def setup(self):
        self.exchange = get_event_exchange(self.container.service_name)
        self.declare.append(self.exchange)
        super(EventDispatcher, self).setup()

    def get_dependency(self, worker_ctx):
        """ Inject a dispatch method onto the service instance
        """

        def dispatch(event_type, event_data):
            extra_headers = encode_to_headers(worker_ctx.context_data)
            self.publisher.publish(
                event_data,
                exchange=self.exchange,
                routing_key=event_type,
                extra_headers=extra_headers
            )

        return dispatch


class EventHandler(Consumer):

    def __init__(self, source_service, event_type, handler_type=SERVICE_POOL,
                 reliable_delivery=True, requeue_on_error=False, **kwargs):
        r"""
        Decorate a method as a handler of ``event_type`` events on the service
        called ``source_service``.
        :Parameters:
            source_service : str
                Name of the service that dispatches the event
            event_type : str
                Type of the event to handle
            handler_type : str
                Determines the behaviour of the handler in a cluster:
                - ``events.SERVICE_POOL``:
                    Event handlers are pooled by service type and method,
                    and one service instance from each pool receives the
                    event. ::
                                   .-[queue]- (service X handler-meth-1)
                                  /
                        exchange o --[queue]- (service X handler-meth-2)
                                  \
                                   \          (service Y(inst. 1) handler-meth)
                                    \       /
                                     [queue]
                                            \
                                              (service Y(inst. 2) handler-meth)
                - ``events.SINGLETON``:
                    Events are received by only one registered handler,
                    regardless of service type. If requeued on error, they may
                    be handled by a different service instance. ::
                                               (service X handler-meth)
                                             /
                        exchange o -- [queue]
                                             \
                                               (service Y handler-meth)
                - ``events.BROADCAST``:
                    Events will be received by every handler. Events are
                    broadcast to every service instance, not just every service
                    type. Instances are differentiated using
                    :attr:`EventHandler.broadcast_identifier`. ::
                                    [queue]- (service X(inst. 1) handler-meth)
                                  /
                        exchange o - [queue]- (service X(inst. 2) handler-meth)
                                  \
                                    [queue]- (service Y handler-meth)

            requeue_on_error : bool  # TODO: defined by Consumer actually..
                If true, handlers will return the event to the queue if an
                error occurs while handling it. Defaults to False.
            reliable_delivery : bool
                If true, events will be held in the queue until there is a
                handler to consume them. Defaults to True.
        """
        self.source_service = source_service
        self.event_type = event_type
        self.handler_type = handler_type
        self.reliable_delivery = reliable_delivery

        super(EventHandler, self).__init__(
            queue=None, requeue_on_error=requeue_on_error, **kwargs
        )

    @property
    def broadcast_identifier(self):
        """ A unique string to identify a service instance for `BROADCAST`
        type handlers.

        The `broadcast_identifier` is appended to the queue name when the
        `BROADCAST` handler type is used. It must uniquely identify service
        instances that receive broadcasts.

        The default `broadcast_identifier` is a uuid that is set when the
        service starts. It will change when the service restarts, meaning
        that any unconsumed messages that were broadcast to the 'old' service
        instance will not be received by the 'new' one. ::

            @property
            def broadcast_identifier(self):
                # use a uuid as the identifier.
                # the identifier will change when the service restarts and
                # any unconsumed messages will be lost
                return uuid.uuid4().hex

        The default behaviour is therefore incompatible with reliable delivery.

        An alternative `broadcast_identifier` that would survive service
        restarts is ::

            @property
            def broadcast_identifier(self):
                # use the machine hostname as the identifier.
                # this assumes that only one instance of a service runs on
                # any given machine
                return socket.gethostname()

        If neither of these approaches are appropriate, you could read the
        value out of a configuration file ::

            fron nameko import config

            @property
            def broadcast_identifier(self):
                return config['SERVICE_IDENTIFIER']  # or similar

        Broadcast queues are exclusive to ensure that `broadcast_identifier`
        values are unique.

        Because this method is a descriptor, it will be called during
        container creation, regardless of the configured `handler_type`.
        See :class:`nameko.extensions.Extension` for more details.
        """
        if self.handler_type is not BROADCAST:
            return None

        if self.reliable_delivery:
            raise EventHandlerConfigurationError(
                "You are using the default broadcast identifier "
                "which is not compatible with reliable delivery. See "
                ":meth:`nameko.events.EventHandler.broadcast_identifier` "
                "for details.")

        return uuid.uuid4().hex

    def setup(self):
        _log.debug('starting %s', self)

        # handler_type determines queue name and exclusive flag
        exclusive = False
        service_name = self.container.service_name
        if self.handler_type is SERVICE_POOL:
            queue_name = "evt-{}-{}--{}.{}".format(self.source_service,
                                                   self.event_type,
                                                   service_name,
                                                   self.method_name)
        elif self.handler_type is SINGLETON:
            queue_name = "evt-{}-{}".format(self.source_service,
                                            self.event_type)
        elif self.handler_type is BROADCAST:
            broadcast_identifier = self.broadcast_identifier
            queue_name = "evt-{}-{}--{}.{}-{}".format(self.source_service,
                                                      self.event_type,
                                                      service_name,
                                                      self.method_name,
                                                      broadcast_identifier)

        exchange = get_event_exchange(self.source_service)

        # queues for handlers without reliable delivery should be marked as
        # auto-delete so they're removed when the consumer disconnects
        auto_delete = self.reliable_delivery is False

        # queues for broadcast handlers are exclusive (meaning that only one
        # consumer may be connected) except when reliable delivery is enabled,
        # because exclusive queues are always removed when the consumer
        # disconnects, regardless of the value of auto_delete
        exclusive = self.handler_type is BROADCAST
        if self.reliable_delivery:
            exclusive = False

        self.queue = Queue(
            queue_name, exchange=exchange, routing_key=self.event_type,
            durable=True, auto_delete=auto_delete, exclusive=exclusive)

        super(EventHandler, self).setup()


event_handler = EventHandler.decorator



from nameko.rpc import ServiceRpc, rpc


class Service:
    name = "service"

    # we depend on the RPC interface of "another_service"
    other_rpc = ServiceRpc("another_service")

    @rpc  # `method` is exposed over RPC
    def method(self):
        # application logic goes here
        pass



import json
import uuid

import pytest
import yaml
from kombu import Exchange, Queue
from mock import Mock, call

from nameko import config
from nameko.constants import (
    ACCEPT_CONFIG_KEY, SERIALIZER_CONFIG_KEY, SERIALIZERS_CONFIG_KEY
)
from nameko.events import EventDispatcher, event_handler
from nameko.exceptions import ConfigurationError, RemoteError
from nameko.messaging import consume
from nameko.rpc import ServiceRpc, rpc
from nameko.standalone.rpc import ServiceRpcClient
from nameko.testing.services import entrypoint_hook, entrypoint_waiter


entrypoint_called = Mock()


test_data = {
    "hello": ("world",),
    123: 456,
    'abc': [7, 8, 9],
    'foobar': 1.5,
}

serialized_info = {
    'json': {
        'content_type': 'application/json',
        'content_encoding': 'utf-8',
        'data': json.loads(json.dumps(test_data))
    },
    'pickle': {
        'content_type': 'application/x-python-serialize',
        'content_encoding': 'binary',
        'data': test_data
    }
}

unserializable = object()


class Service(object):
    name = "service"

    @rpc
    def echo(self, arg):
        entrypoint_called(arg)
        return arg

    @rpc
    def broken(self):
        return unserializable


@pytest.fixture
def sniffer_queue_factory(rabbit_config, rabbit_manager, get_vhost):
    """ Return a function that creates message queues to 'sniff' messages
    published to exchanges.
    """
    vhost = get_vhost(config['AMQP_URI'])

    def make(exchange, routing_key="*"):
        """ Create a uniquely named queue and bind it to an exchange so that
        it collects messages published to that exchange.

        :Parameters:
            exchange : str
                Name of the exchange to bind to
            routing_key : str
                Routing key to bind with

        :Returns:
            A function that returns all the messages received by the queue

        """
        queue_name = "sniffer_{}".format(uuid.uuid4())
        rabbit_manager.create_queue(vhost, queue_name, auto_delete=True)
        rabbit_manager.create_queue_binding(
            vhost, exchange, queue_name, routing_key=routing_key)

        def get_messages():
            """ Return all messages received by the sniffer queue and remove
            them from the queue.
            """
            return rabbit_manager.get_messages(vhost, queue_name)
        return get_messages

    return make


@pytest.mark.usefixtures("rabbit_config")
@pytest.mark.parametrize("serializer", ['json', 'pickle'])
def test_rpc_serialization(container_factory, sniffer_queue_factory, serializer):

    with config.patch({SERIALIZER_CONFIG_KEY: serializer}):
        container = container_factory(Service)
        container.start()

        get_messages = sniffer_queue_factory('nameko-rpc')

        serialized = serialized_info[serializer]

        with ServiceRpcClient('service') as client:
            assert client.echo(test_data) == serialized['data']
            assert entrypoint_called.call_args == call(serialized['data'])

    msg = get_messages()[0]
    assert msg['properties']['content_type'] == serialized['content_type']


@pytest.mark.usefixtures("rabbit_config")
def test_rpc_result_serialization_error(container_factory):

    container = container_factory(Service)
    container.start()

    with ServiceRpcClient('service') as client:
        with pytest.raises(RemoteError) as exc:
            client.broken()
        assert exc.value.exc_type == "UnserializableValueError"

        assert client.echo('foo') == "foo"  # subsequent calls ok


@pytest.mark.usefixtures("rabbit_config")
def test_rpc_arg_serialization_error(container_factory):

    container = container_factory(Service)
    container.start()

    with ServiceRpcClient('service') as client:
        with pytest.raises(Exception):
            client.echo(unserializable)

        assert client.echo('foo') == "foo"  # subsequent calls ok


@pytest.mark.usefixtures("rabbit_config")
@pytest.mark.parametrize("serializer", ['json', 'pickle'])
def test_event_serialization(
    container_factory, sniffer_queue_factory, serializer
):
    handler_called = Mock()

    class Service(object):
        name = "srcservice"
        dispatch = EventDispatcher()

        @rpc
        def dispatch_event(self, payload):
            self.dispatch("eventtype", payload)

        @event_handler('srcservice', 'eventtype')
        def handle_event(self, event_data):
            handler_called(event_data)

    with config.patch({SERIALIZER_CONFIG_KEY: serializer}):

        container = container_factory(Service)
        container.start()

        get_messages = sniffer_queue_factory(
            "srcservice.events", routing_key="eventtype")

        serialized = serialized_info[serializer]

        # dispatch an event with a tuple payload
        with entrypoint_waiter(container, "handle_event"):
            with entrypoint_hook(container, "dispatch_event") as dispatch_event:
                dispatch_event(test_data)

    # verify data serialized to expected value
    assert handler_called.call_args == call(serialized['data'])

    # verify sniffed messages serialized as expected
    msg = get_messages()[0]
    assert msg['properties']['content_type'] == serialized['content_type']


def upperjson_encode(value):
    value = json.dumps(value)
    return value.upper()


def upperjson_decode(value):
    value = value.lower()
    return json.loads(value)


@pytest.mark.usefixtures("rabbit_config")
def test_custom_serializer(container_factory, sniffer_queue_factory):

    class Service(object):
        name = "service"

        @rpc
        def echo(self, arg):
            return arg

    with config.patch({
        SERIALIZER_CONFIG_KEY: "upperjson",
        SERIALIZERS_CONFIG_KEY: {
            'upperjson': {
                'encoder': 'test.test_serialization.upperjson_encode',
                'decoder': 'test.test_serialization.upperjson_decode',
                'content_type': 'application/x-upper-json'
            }
        }
    }):
        container = container_factory(Service)
        container.start()

        get_messages = sniffer_queue_factory('nameko-rpc')

        # verify RPC works end-to-end
        with ServiceRpcClient('service') as client:
            assert client.echo("hello") == "hello"

    # verify sniffed messages serialized as expected
    msg = get_messages()[0]
    assert '"RESULT": "HELLO"' in msg['payload']
    assert msg['properties']['content_type'] == "application/x-upper-json"


@pytest.mark.usefixtures("rabbit_config")
@pytest.mark.parametrize(
    'cfg',
    (
        {SERIALIZER_CONFIG_KEY: 'unknown'},
        {ACCEPT_CONFIG_KEY: ['json', 'unknown']},
        {
            SERIALIZER_CONFIG_KEY: 'json',
            ACCEPT_CONFIG_KEY: ['json', 'unknown'],
        },
    )
)
def test_missing_serializers(container_factory, cfg):

    with config.patch(cfg):
        with pytest.raises(ConfigurationError) as exc:
            container_factory(Service)

    assert (
        str(exc.value) ==
        'Please register a serializer for "unknown" format')


@pytest.yield_fixture
def multi_serializer_config():
    with config.patch({ACCEPT_CONFIG_KEY: ['json', 'yaml']}):
        yield


@pytest.mark.usefixtures("rabbit_config")
@pytest.mark.usefixtures("multi_serializer_config")
@pytest.mark.parametrize(
    'content_type, encode',
    (
        ('application/json', json.dumps),
        ('application/x-yaml', yaml.dump),
    )
)
def test_consumer_accepts_multiple_serialization_formats(
    container_factory, rabbit_manager, content_type, encode, get_vhost
):

    class Service(object):

        name = 'service'

        @consume(queue=Queue(exchange=Exchange('spam'), name='some-queue'))
        def consume(self, payload):
            assert payload == {'spam': 'ham'}

    container = container_factory(Service)
    container.start()

    payload = {'spam': 'ham'}

    with entrypoint_waiter(container, 'consume'):
        rabbit_manager.publish(
            get_vhost(config['AMQP_URI']), 'spam', '', encode(payload),
            properties={'content_type': content_type})


@pytest.mark.usefixtures("rabbit_config")
@pytest.mark.usefixtures("multi_serializer_config")
@pytest.mark.parametrize(
    'serializer, content_type, encode',
    (
        ('json', 'application/json', json.dumps),
        ('yaml', 'application/x-yaml', yaml.dump),
    )
)
def test_standalone_rpc_accepts_multiple_serialization_formats(
    container_factory, sniffer_queue_factory, serializer,
    content_type, encode
):

    called = Mock()

    class EchoingService(object):

        name = 'echoer'

        @rpc
        def echo(self, payload):
            called(payload)
            return payload

    echoer = container_factory(EchoingService)
    echoer.start()

    payload = {'spam': 'ham'}

    get_messages = sniffer_queue_factory('nameko-rpc')

    with ServiceRpcClient('echoer', serializer=serializer) as client:
        assert client.echo(payload) == payload

    msg = get_messages().pop()
    assert encode(payload) in msg['payload']
    assert msg['properties']['content_type'] == content_type

    assert called.mock_calls == [call(payload)]


@pytest.mark.usefixtures("rabbit_config")
@pytest.mark.usefixtures("multi_serializer_config")
@pytest.mark.parametrize(
    'serializer, content_type, encode',
    (
        ('json', 'application/json', json.dumps),
        ('yaml', 'application/x-yaml', yaml.dump),
    )
)
def test_rpc_accepts_multiple_serialization_formats(
    container_factory, rabbit_manager, sniffer_queue_factory,
    serializer, content_type, encode
):

    called = Mock()

    class ForwardingService(object):

        name = 'forwarder'

        echoer = ServiceRpc('echoer', serializer=serializer)

        @rpc
        def forward(self, payload):
            return self.echoer.echo(payload)

    class EchoingService(object):

        name = 'echoer'

        @rpc
        def echo(self, payload):
            called(payload)
            return payload

    echoer = container_factory(EchoingService)
    echoer.start()

    payload = {'spam': 'ham'}

    forwarder = container_factory(ForwardingService)
    forwarder.start()

    get_messages = sniffer_queue_factory('nameko-rpc')

    with entrypoint_hook(forwarder, 'forward') as echo:
        assert echo(payload) == payload

    msg = get_messages().pop()
    assert encode(payload) in msg['payload']
    assert msg['properties']['content_type'] == content_type

    assert called.mock_calls == [call(payload)]


@pytest.mark.usefixtures("rabbit_config")
@pytest.mark.usefixtures("multi_serializer_config")
@pytest.mark.parametrize(
    'serializer, content_type, encode',
    (
        ('json', 'application/json', json.dumps),
        ('yaml', 'application/x-yaml', yaml.dump),
    )
)
def test_events_accepts_multiple_serialization_formats(
    container_factory, rabbit_manager, sniffer_queue_factory,
    serializer, content_type, encode
):

    called = Mock()

    class PublishingService(object):

        name = 'publisher'

        dispatch = EventDispatcher(serializer=serializer)

        @rpc
        def dispatch_event(self, payload):
            self.dispatch('spam', payload)

    class ConsumingService(object):

        name = 'consumer'

        @event_handler('publisher', 'spam')
        def handle_event(self, event_data):
            called(event_data)

    def publish(consumer, publisher):
        with entrypoint_waiter(consumer, "handle_event"):
            with entrypoint_hook(publisher, "dispatch_event") as dispatch:
                dispatch(payload)

    consumer = container_factory(ConsumingService)
    consumer.start()

    get_messages = sniffer_queue_factory(
        'publisher.events', routing_key='spam')

    payload = {'spam': 'ham'}

    publisher = container_factory(PublishingService)
    publisher.start()

    publish(consumer, publisher)

    msg = get_messages().pop()
    assert encode(payload) in msg['payload']
    assert msg['properties']['content_type'] == content_type

    assert called.mock_calls == [call(payload)]



import socket

import pytest
from eventlet import wsgi
from mock import patch

from nameko.exceptions import ConfigurationError
from nameko.web.handlers import HttpRequestHandler, http
from nameko.web.server import (
    BaseHTTPServer, HttpOnlyProtocol, WebServer, parse_address
)


try:
    from werkzeug.middleware.proxy_fix import ProxyFix
except ImportError:  # pragma: no cover
    from werkzeug.contrib.fixers import ProxyFix  # werkzeug<1


class ExampleService(object):
    name = "exampleservice"

    @http('GET', '/')
    def do_index(self, request):
        return ''

    @http('GET', '/large')
    def do_large(self, request):
        # more than a buffer's worth
        return 'x' * (10**6)


@pytest.mark.usefixtures("web_config")
def test_broken_pipe(container_factory, web_config_port, web_session):
    container = container_factory(ExampleService)
    container.start()

    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    s.connect(('127.0.0.1', web_config_port))
    s.sendall(b'GET /large \r\n\r\n')
    s.recv(10)
    s.close()  # break connection while there is still more data coming

    # server should still work
    assert web_session.get('/').text == ''


@pytest.mark.usefixtures("web_config")
def test_other_socket_error(container_factory, web_config_port, web_session):
    container = container_factory(ExampleService)
    container.start()

    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    s.connect(('127.0.0.1', web_config_port))

    with patch.object(BaseHTTPServer.BaseHTTPRequestHandler, 'finish') as fin:
        fin.side_effect = socket.error('boom')
        s.sendall(b'GET / \r\n\r\n')
        s.recv(10)
        s.close()

    # takes down container
    with pytest.raises(socket.error) as exc:
        container.wait()
    assert 'boom' in str(exc)


@pytest.mark.usefixtures("web_config")
def test_client_disconnect_os_error(
    container_factory, web_config_port, web_session
):
    """ Regression for https://github.com/nameko/nameko/issues/368
    """
    container = container_factory(ExampleService)
    container.start()

    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    s.connect(('127.0.0.1', web_config_port))

    with patch.object(HttpOnlyProtocol, 'handle_one_request') as handle:
        handle.side_effect = OSError('raw readinto() returned invalid length')
        s.sendall(b'GET / \r\n\r\n')
        s.recv(10)
        s.close()

    # server should still work
    assert web_session.get('/').text == ''


@pytest.mark.usefixtures("web_config")
def test_other_os_error(container_factory, web_config_port, web_session):
    """ Regression for https://github.com/nameko/nameko/issues/368
    """
    container = container_factory(ExampleService)
    container.start()

    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    s.connect(('127.0.0.1', web_config_port))

    with patch.object(BaseHTTPServer.BaseHTTPRequestHandler, 'finish') as fin:
        fin.side_effect = OSError('boom')
        s.sendall(b'GET / \r\n\r\n')
        s.recv(10)
        s.close()

    # takes down container
    with pytest.raises(OSError) as exc:
        container.wait()
    assert 'boom' in str(exc)


@pytest.mark.parametrize(['source', 'result'], [
    ('8000', ('', 8000)),
    ('foo:8000', ('foo', 8000)),
    ('foo', None),
])
def test_parse_address(source, result):
    if result is None:
        with pytest.raises(ConfigurationError) as exc:
            parse_address(source)
        assert 'Misconfigured bind address' in str(exc)
        assert '`foo`' in str(exc)

    else:
        assert parse_address(source) == result


@pytest.mark.usefixtures("web_config")
def test_adding_middleware_with_get_wsgi_app(container_factory):

    class CustomWebServer(WebServer):
        def get_wsgi_app(self):
            # get the original WSGI app that processes http requests
            app = super(CustomWebServer, self).get_wsgi_app()
            # apply the ProxyFix middleware as an example
            return ProxyFix(app)

    class CustomHttpRequestHandler(HttpRequestHandler):
        server = CustomWebServer()

    http = CustomHttpRequestHandler.decorator

    class CustomServerExampleService(object):
        name = 'customserverservice'

        @http('GET', '/')
        def do_index(self, request):
            return ''  # pragma: no cover

    container = container_factory(CustomServerExampleService)
    with patch.object(CustomWebServer, 'get_wsgi_server') as get_wsgi_server:
        container.start()

    wsgi_app = get_wsgi_server.call_args[0][1]
    assert isinstance(wsgi_app, ProxyFix)


@pytest.mark.usefixtures("web_config")
def test_custom_wsgi_server_is_used(
    container_factory, web_config_port, web_session
):
    def custom_wsgi_app(environ, start_response):
        start_response('200 OK', [])
        return 'Override'

    class CustomWebServer(WebServer):
        def get_wsgi_server(
            self, sock, wsgi_app, protocol=HttpOnlyProtocol, debug=False
        ):
            return wsgi.Server(
                sock,
                sock.getsockname(),
                custom_wsgi_app,
                protocol=protocol,
                debug=debug
            )

    class CustomHttpRequestHandler(HttpRequestHandler):
        server = CustomWebServer()

    http = CustomHttpRequestHandler.decorator

    class CustomServerExampleService(object):
        name = 'customserverservice'

        @http('GET', '/')
        def do_index(self, request):
            return ''  # pragma: no cover

    container = container_factory(CustomServerExampleService)
    container.start()

    assert web_session.get('/').text == 'Override'



import os
import sys
from types import ModuleType

from nameko import config
from nameko.constants import AMQP_URI_CONFIG_KEY
from nameko.standalone.events import event_dispatcher
from nameko.standalone.rpc import ClusterRpcClient

from .utils import interact


SHELLS = ["bpython", "ipython", "plain"]


class ShellRunner(object):
    def __init__(self, banner, local):
        self.banner = banner
        self.local = local

    def bpython(self):
        import bpython  # pylint: disable=E0401

        bpython.embed(banner=self.banner, locals_=self.local)

    def ipython(self):
        from IPython import embed  # pylint: disable=E0401

        embed(banner1=self.banner, user_ns=self.local)

    def plain(self):
        interact(
            banner=self.banner,
            local=self.local,
            raise_expections=not sys.stdin.isatty(),
        )

    def start_shell(self, name):
        if not sys.stdin.isatty():
            available_shells = ["plain"]
        else:
            available_shells = [name] if name else SHELLS

        # Support the regular Python interpreter startup script if someone
        # is using it.
        startup = os.environ.get("PYTHONSTARTUP")
        if startup and os.path.isfile(startup):
            with open(startup, "r") as f:
                eval(compile(f.read(), startup, "exec"), self.local)
            del os.environ["PYTHONSTARTUP"]

        for name in available_shells:
            try:
                return getattr(self, name)()
            except ImportError:  # pragma: no cover noqa: F401
                pass
        self.plain()  # pragma: no cover


def make_nameko_helper():
    """Create a fake module that provides some convenient access to nameko
    standalone functionality for interactive shell usage.
    """
    module = ModuleType("nameko")
    module.__doc__ = """Nameko shell helper for making rpc calls and dispatching
events.

Usage:
    >>> n.rpc.service.method()
    "reply"

    >>> n.dispatch_event('service', 'event_type', 'event_data')
"""
    client = ClusterRpcClient()
    module.rpc = client.start()
    module.dispatch_event = event_dispatcher()
    module.config = config
    module.disconnect = client.stop
    return module


def main(interface):

    banner = "Nameko Python %s shell on %s\nBroker: %s" % (
        sys.version,
        sys.platform,
        config[AMQP_URI_CONFIG_KEY],
    )

    ctx = {}
    ctx["n"] = make_nameko_helper()

    runner = ShellRunner(banner, ctx)
    runner.start_shell(name=interface)



import itertools
import time

import pytest
from kombu.messaging import Queue
from mock import Mock

from nameko import config
from nameko.amqp.publish import Publisher
from nameko.constants import DEFAULT_HEARTBEAT, HEARTBEAT_CONFIG_KEY
from nameko.events import EventHandler, event_handler
from nameko.messaging import Consumer, consume
from nameko.rpc import ReplyListener, RpcConsumer, ServiceRpc, rpc
from nameko.standalone.events import event_dispatcher
from nameko.standalone.rpc import ServiceRpcClient
from nameko.testing.services import entrypoint_waiter, once
from nameko.testing.utils import get_extension


class TestDeadlockRegression(object):
    """ Regression test for https://github.com/nameko/nameko/issues/428
    """

    @pytest.yield_fixture
    def config(self, rabbit_config):
        with config.patch({'max_workers': 2}):
            yield

    @pytest.fixture
    def upstream(self, container_factory, config):

        class Service(object):
            name = "upstream"

            @rpc
            def method(self):
                time.sleep(.5)

        container = container_factory(Service)
        container.start()

    @pytest.fixture
    def service_cls(self):

        class Service(object):
            name = "downsteam"

            upstream_rpc = ServiceRpc("upstream")

            @event_handler('service', 'event1')
            def handle_event1(self, event_data):
                self.upstream_rpc.method()

            @event_handler('service', 'event2')
            def handle_event2(self, event_data):
                self.upstream_rpc.method()

        return Service

    @pytest.mark.usefixtures('config')
    @pytest.mark.usefixtures('upstream')
    def test_deadlock_due_to_slow_workers(
        self, service_cls, container_factory
    ):
        """
        Implementation has now changed, but keeping this test as a regression.

        Deadlock would occur if the unack'd messages grows beyond the
        size of the worker pool at any point. The QueueConsumer would block
        waiting for a worker and pending RPC replies would not be ack'd.
        Any running workers therefore never complete, and the worker pool
        remains exhausted.
        """
        container = container_factory(service_cls)
        container.start()

        count = 2

        dispatch = event_dispatcher()
        for _ in range(count):
            dispatch("service", "event1", 1)
            dispatch("service", "event2", 1)

        counter = itertools.count(start=1)

        def cb(worker_ctx, res, exc_info):
            if next(counter) == count:
                return True

        with entrypoint_waiter(
            container, 'handle_event1', timeout=5, callback=cb
        ):
            pass


class TestHeartbeats(object):

    @pytest.fixture
    def service_cls(self):
        class Service(object):
            name = "service"

            upstream_rpc = ServiceRpc("upstream")

            @consume(queue=Queue(name="queue"))
            @event_handler("service", "event")
            @rpc
            def echo(self, arg):
                return arg  # pragma: no cover

        return Service

    @pytest.mark.usefixtures('rabbit_config')
    @pytest.mark.parametrize(
        'extension_cls', [Consumer, EventHandler, RpcConsumer, ReplyListener]
    )
    def test_default(self, extension_cls, service_cls, container_factory):

        container = container_factory(service_cls)
        container.start()

        extension = get_extension(container, extension_cls)
        assert extension.consumer.heartbeat == DEFAULT_HEARTBEAT

    @pytest.mark.usefixtures('rabbit_config')
    @pytest.mark.parametrize("heartbeat", [30, None])
    @pytest.mark.parametrize(
        'extension_cls', [Consumer, EventHandler, RpcConsumer, ReplyListener]
    )
    def test_config_value(
        self, extension_cls, heartbeat, service_cls, container_factory
    ):
        with config.patch({HEARTBEAT_CONFIG_KEY: heartbeat}):
            container = container_factory(service_cls)
            container.start()

            extension = get_extension(container, extension_cls)
            assert extension.consumer.heartbeat == heartbeat


class TestHeartbeatFailure(object):

    @pytest.fixture
    def config(self, rabbit_config):
        with config.patch({
            'max_workers': 2,
            HEARTBEAT_CONFIG_KEY: 3  # minimum reliable heartbeat
        }):
            yield

    @pytest.fixture
    def tracker(self):
        return Mock()

    @pytest.mark.usefixtures('config')
    def test_event_handler(
        self, container_factory, tracker, queue_info
    ):

        class Service(object):
            name = "service"

            @once
            def work(self):
                # consume a worker
                time.sleep(999)

            @event_handler('service', 'event')
            def handle(self, event_data):
                time.sleep(9)  # >2x heartbeat
                tracker(event_data)

        container = container_factory(Service)
        container.start()

        # wait for service to handle two requests
        counter = itertools.count(start=1)
        with entrypoint_waiter(
            container, 'handle', callback=lambda *a: next(counter) == 2
        ):
            # dispatch two messages; the second will block on the worker pool
            dispatch = event_dispatcher()
            dispatch("service", "event", 1)
            dispatch("service", "event", 2)

        # expect entrypoint to only have run twice and consumed all messages.
        # more than twice or remaining messages means the heartbeat failed
        # and a message was requeued or redelivered
        assert tracker.call_count == 2
        assert queue_info(
            "evt-service-event--service.handle"
        ).message_count == 0

    @pytest.mark.usefixtures('config')
    def test_consumer(self, container_factory, tracker, queue_info):

        queue = Queue(name="queue")

        class Service(object):
            name = "service"

            @once
            def work(self):
                # consume a worker
                time.sleep(999)

            @consume(queue)
            def handle(self, payload):
                time.sleep(9)  # >3x heartbeat
                tracker(payload)

        container = container_factory(Service)
        container.start()

        # wait for service to handle two requests
        counter = itertools.count(start=1)
        with entrypoint_waiter(
            container, 'handle', callback=lambda *a: next(counter) == 2
        ):
            # publish two messages; the second will block on the worker pool
            publisher = Publisher(
                config['AMQP_URI'], routing_key=queue.name
            )
            publisher.publish(1)
            publisher.publish(2)

        # expect entrypoint to only have run twice and consumed all messages.
        # more than twice or remaining messages means the heartbeat failed
        # and a message was requeued or redelivered
        assert tracker.call_count == 2
        assert queue_info(queue.name).message_count == 0

    @pytest.mark.usefixtures('config')
    def test_rpc(self, container_factory, tracker, queue_info):

        class Service(object):
            name = "service"

            @once
            def work(self):
                # consume a worker
                time.sleep(999)

            @rpc
            def method(self, arg):
                time.sleep(9)  # >2x heartbeat
                tracker(arg)

        container = container_factory(Service)
        container.start()

        # wait for service to handle two requests
        counter = itertools.count(start=1)
        with entrypoint_waiter(
            container, 'method', callback=lambda *a: next(counter) == 2
        ):
            # make two requests; the second will block on the worker pool
            with ServiceRpcClient('service') as service_rpc:
                service_rpc.method.call_async(1)
                service_rpc.method.call_async(2)

        # expect entrypoint to only have run twice and consumed all messages.
        # more than twice or remaining messages means the heartbeat failed
        # and a message was requeued or redelivered
        assert tracker.call_count == 2
        assert queue_info("rpc-service").message_count == 0



from kombu import Exchange

from nameko import config, serialization
from nameko.amqp.publish import Publisher
from nameko.constants import (
    AMQP_SSL_CONFIG_KEY, AMQP_URI_CONFIG_KEY, DEFAULT_AMQP_URI, PERSISTENT
)


def get_event_exchange(service_name):
    """ Get an exchange for ``service_name`` events.
    """
    exchange_name = "{}.events".format(service_name)
    exchange = Exchange(
        exchange_name, type='topic', durable=True, delivery_mode=PERSISTENT)

    return exchange


def event_dispatcher(**kwargs):
    """ Return a function that dispatches nameko events.
    """
    amqp_uri = config.get(AMQP_URI_CONFIG_KEY, DEFAULT_AMQP_URI)
    amqp_uri = kwargs.pop('uri', amqp_uri)

    serializer, _ = serialization.setup()
    serializer = kwargs.pop('serializer', serializer)

    ssl = config.get(AMQP_SSL_CONFIG_KEY)
    ssl = kwargs.pop('ssl', ssl)

    # TODO: standalone event dispatcher should accept context event_data
    # and insert a call id

    publisher = Publisher(amqp_uri, serializer=serializer, ssl=ssl, **kwargs)

    def dispatch(service_name, event_type, event_data):
        """ Dispatch an event claiming to originate from `service_name` with
        the given `event_type` and `event_data`.
        """
        exchange = get_event_exchange(service_name)

        publisher.publish(
            event_data,
            exchange=exchange,
            routing_key=event_type
        )

    return dispatch



"""Load configuration from YAML file. Allow env var rendering.

E.g.::

    fname = "conf.yaml"
    define = {"MY_PARAM": "is here", "YOUR_PARAM": "is here too"}
    with open(fname, "rb") as config_file:
        setup_config(config_file)
        # from now on, nameko.config is updated
    from nameko import config
    assert "MY_PARAM" in config
    assert "YOUR_PARAM" in config
"""
from __future__ import print_function

import os
import re
import warnings
from functools import partial

import yaml

from nameko import config
from nameko.constants import AMQP_URI_CONFIG_KEY


try:
    import regex
except ImportError:  # pragma: no cover
    ENV_VAR_MATCHER = re.compile(
        r"""
            \$\{       # match characters `${` literally
            ([^}:\s]+) # 1st group: matches any character except `}` or `:`
            :?         # matches the literal `:` character zero or one times
            ([^}]+)?   # 2nd group: matches any character except `}`
            \}         # match character `}` literally
        """,
        re.VERBOSE,
    )
else:  # pragma: no cover
    ENV_VAR_MATCHER = regex.compile(
        r"""
        \$\{                #  match ${
        (                   #  first capturing group: variable name
            [^{}:\s]+       #  variable name without {,},: or spaces
        )
        (?:                 # non capturing optional group for value
            :               # match :
            (               # 2nd capturing group: default value
                (?:         # non capturing group for OR
                    [^{}]   # any non bracket
                |           # OR
                    \{      # literal {
                    (?2)    # recursive 2nd capturing group aka ([^{}]|{(?2)})
                    \}      # literal }
                )*          #
            )
        )?
        \}                  # end of macher }
        """,
        regex.VERBOSE,
    )

IMPLICIT_ENV_VAR_MATCHER = re.compile(
    r"""
        .*          # matches any number of any characters
        \$\{.*\}    # matches any number of any characters
                    # between `${` and `}` literally
        .*          # matches any number of any characters
    """,
    re.VERBOSE,
)


def _replace_env_var(match):
    env_var, default = match.groups()
    value = os.environ.get(env_var, None)
    if value is None:
        # expand default using other vars
        if default is None:
            # regex module return None instead of
            #  '' if engine didn't entered default capture group
            default = ""

        value = default
        while IMPLICIT_ENV_VAR_MATCHER.match(value):  # pragma: no cover
            value = ENV_VAR_MATCHER.sub(_replace_env_var, value)
    return value


def env_var_constructor(loader, node, raw=False):
    raw_value = loader.construct_scalar(node)
    value = ENV_VAR_MATCHER.sub(_replace_env_var, raw_value)
    return value if raw else yaml.safe_load(value)


def setup_yaml_parser():
    yaml.add_constructor("!env_var", env_var_constructor, yaml.UnsafeLoader)
    yaml.add_constructor(
        "!raw_env_var", partial(env_var_constructor, raw=True), yaml.UnsafeLoader
    )
    yaml.add_implicit_resolver(
        "!env_var", IMPLICIT_ENV_VAR_MATCHER, Loader=yaml.UnsafeLoader
    )


def load_config(config_file):
    setup_yaml_parser()
    return yaml.unsafe_load(config_file)


def setup_config(config_file, define=None, broker=None):
    if config_file:
        config.update(load_config(config_file))
    # --broker is deprecated, emit warning if used
    if broker:
        warnings.warn(
            (
                "--broker option is going to be removed. ",
                "Use --config or --define and set AMQP_URI instead.",
            ),
            DeprecationWarning,
        )
        if AMQP_URI_CONFIG_KEY not in config:
            config.update({AMQP_URI_CONFIG_KEY: broker})
    if define:
        config.update(define)
    return config



"""
This file defines several toy services that interact to form a shop of the
famous ACME Corporation. The AcmeShopService relies on the StockService,
InvoiceService and PaymentService to fulfil its orders. They are not best
practice examples! They're minimal services provided for the test at the
bottom of the file.

``test_shop_integration`` is a full integration test of the ACME shop
"checkout flow". It demonstrates how to test the multiple ACME services in
combination with each other, including limiting service interactions by
replacing certain entrypoints and dependencies.
"""

from collections import defaultdict

import pytest

from nameko.events import EventDispatcher, event_handler
from nameko.exceptions import RemoteError
from nameko.extensions import DependencyProvider
from nameko.rpc import ServiceRpc, rpc
from nameko.standalone.rpc import ServiceRpcClient
from nameko.testing.services import replace_dependencies, restrict_entrypoints
from nameko.testing.utils import get_container
from nameko.timer import timer


class NotLoggedInError(Exception):
    pass


class ItemOutOfStockError(Exception):
    pass


class ItemDoesNotExistError(Exception):
    pass


class ShoppingBasket(DependencyProvider):
    """ A shopping basket tied to the current ``user_id``.
    """
    def __init__(self):
        self.baskets = defaultdict(list)

    def get_dependency(self, worker_ctx):

        class Basket(object):
            def __init__(self, basket):
                self._basket = basket
                self.worker_ctx = worker_ctx

            def add(self, item):
                self._basket.append(item)

            def __iter__(self):
                for item in self._basket:
                    yield item

        try:
            user_id = worker_ctx.data['user_id']
        except KeyError:
            raise NotLoggedInError()
        return Basket(self.baskets[user_id])


class AcmeShopService:
    name = "acmeshopservice"

    user_basket = ShoppingBasket()
    stock_rpc = ServiceRpc('stockservice')
    invoice_rpc = ServiceRpc('invoiceservice')
    payment_rpc = ServiceRpc('paymentservice')

    fire_event = EventDispatcher()

    @rpc
    def add_to_basket(self, item_code):
        """ Add item identified by ``item_code`` to the shopping basket.

        This is a toy example! Ignore the obvious race condition.
        """
        stock_level = self.stock_rpc.check_stock(item_code)
        if stock_level > 0:
            self.user_basket.add(item_code)
            self.fire_event("item_added_to_basket", item_code)
            return item_code

        raise ItemOutOfStockError(item_code)

    @rpc
    def checkout(self):
        """ Take payment for all items in the shopping basket.
        """
        total_price = sum(self.stock_rpc.check_price(item)
                          for item in self.user_basket)

        # prepare invoice
        invoice = self.invoice_rpc.prepare_invoice(total_price)

        # take payment
        self.payment_rpc.take_payment(invoice)

        # fire checkout event if prepare_invoice and take_payment succeeded
        checkout_event_data = {
            'invoice': invoice,
            'items': list(self.user_basket)
        }
        self.fire_event("checkout_complete", checkout_event_data)
        return total_price


class Warehouse(DependencyProvider):
    """ A database of items in the warehouse.

    This is a toy example! A dictionary is not a database.
    """
    def __init__(self):
        self.database = {
            'anvil': {
                'price': 100,
                'stock': 3
            },
            'dehydrated_boulders': {
                'price': 999,
                'stock': 12
            },
            'invisible_paint': {
                'price': 10,
                'stock': 30
            },
            'toothpicks': {
                'price': 1,
                'stock': 0
            }
        }

    def get_dependency(self, worker_ctx):
        return self.database


class StockService:
    name = "stockservice"

    warehouse = Warehouse()

    @rpc
    def check_price(self, item_code):
        """ Check the price of an item.
        """
        try:
            return self.warehouse[item_code]['price']
        except KeyError:
            raise ItemDoesNotExistError(item_code)

    @rpc
    def check_stock(self, item_code):
        """ Check the stock level of an item.
        """
        try:
            return self.warehouse[item_code]['stock']
        except KeyError:
            raise ItemDoesNotExistError(item_code)

    @rpc
    @timer(100)
    def monitor_stock(self):
        """ Periodic stock monitoring method. Can also be triggered manually
        over RPC.

        This is an expensive process that we don't want to exercise during
        integration testing...
        """
        raise NotImplementedError()

    @event_handler('acmeshopservice', "checkout_complete")
    def dispatch_items(self, event_data):
        """ Dispatch items from stock on successful checkouts.

        This is an expensive process that we don't want to exercise during
        integration testing...
        """
        raise NotImplementedError()


class AddressBook(DependencyProvider):
    """ A database of user details, keyed on user_id.
    """
    def __init__(self):
        self.address_book = {
            'wile_e_coyote': {
                'username': 'wile_e_coyote',
                'fullname': 'Wile E Coyote',
                'address': '12 Long Road, High Cliffs, Utah',
            },
        }

    def get_dependency(self, worker_ctx):
        def get_user_details():
            try:
                user_id = worker_ctx.data['user_id']
            except KeyError:
                raise NotLoggedInError()
            return self.address_book.get(user_id)
        return get_user_details


class InvoiceService:
    name = "invoiceservice"

    get_user_details = AddressBook()

    @rpc
    def prepare_invoice(self, amount):
        """ Prepare an invoice for ``amount`` for the current user.
        """
        address = self.get_user_details().get('address')
        fullname = self.get_user_details().get('fullname')
        username = self.get_user_details().get('username')

        msg = "Dear {}. Please pay ${} to ACME Corp.".format(fullname, amount)
        invoice = {
            'message': msg,
            'amount': amount,
            'customer': username,
            'address': address
        }
        return invoice


class PaymentService:
    name = "paymentservice"

    @rpc
    def take_payment(self, invoice):
        """ Take payment from a customer according to ``invoice``.

        This is an expensive process that we don't want to exercise during
        integration testing...
        """
        raise NotImplementedError()

# =============================================================================
# Begin test
# =============================================================================


@pytest.yield_fixture
def rpc_client_factory(rabbit_config):
    """ Factory fixture for standalone RPC clients.

    Clients are started automatically so they can be used without a ``with``
    statement. All created clients are stopped at the end of the test, when
    this fixture closes.
    """
    all_clients = []

    def make_client(service_name, **kwargs):
        client = ServiceRpcClient(service_name, **kwargs)
        all_clients.append(client)
        return client.start()

    yield make_client

    for client in all_clients:
        client.stop()


def test_shop_checkout_integration(runner_factory, rpc_client_factory):
    """ Simulate a checkout flow as an integration test.

    Requires instances of AcmeShopService, StockService and InvoiceService
    to be running. Explicitly replaces the rpc client to PaymentService so
    that service doesn't need to be hosted.

    Also replaces the event dispatcher dependency on AcmeShopService and
    disables the timer entrypoint on StockService. Limiting the interactions
    of services in this way reduces the scope of the integration test and
    eliminates undesirable side-effects (e.g. processing events unnecessarily).
    """
    context_data = {'user_id': 'wile_e_coyote'}
    shop = rpc_client_factory('acmeshopservice', context_data=context_data)

    runner = runner_factory(AcmeShopService, StockService, InvoiceService)

    # replace ``event_dispatcher`` and ``payment_rpc``  dependencies on
    # AcmeShopService with ``MockDependencyProvider``\s
    shop_container = get_container(runner, AcmeShopService)
    fire_event, payment_rpc = replace_dependencies(
        shop_container, "fire_event", "payment_rpc")

    # restrict entrypoints on StockService
    stock_container = get_container(runner, StockService)
    restrict_entrypoints(stock_container, "check_price", "check_stock")

    runner.start()

    # add some items to the basket
    assert shop.add_to_basket("anvil") == "anvil"
    assert shop.add_to_basket("invisible_paint") == "invisible_paint"

    # try to buy something that's out of stock
    with pytest.raises(RemoteError) as exc_info:
        shop.add_to_basket("toothpicks")
    assert exc_info.value.exc_type == "ItemOutOfStockError"

    # provide a mock response from the payment service
    payment_rpc.take_payment.return_value = "Payment complete."

    # checkout
    res = shop.checkout()

    total_amount = 100 + 10
    assert res == total_amount

    # verify integration with mocked out payment service
    payment_rpc.take_payment.assert_called_once_with({
        'customer': "wile_e_coyote",
        'address': "12 Long Road, High Cliffs, Utah",
        'amount': total_amount,
        'message': "Dear Wile E Coyote. Please pay $110 to ACME Corp."
    })

    # verify events fired as expected
    assert fire_event.call_count == 3


if __name__ == "__main__":
    import sys
    pytest.main(sys.argv)



import itertools

import eventlet
import pytest
from eventlet.event import Event
from mock import Mock, call, patch
from requests import HTTPError, Response

from nameko.containers import ServiceContainer
from nameko.rpc import Rpc, RpcConsumer, rpc
from nameko.testing.rabbit import Client
from nameko.testing.utils import (
    AnyInstanceOf, ResourcePipeline, find_free_port, get_container,
    get_extension, get_rabbit_connections, reset_rabbit_connections,
    wait_for_call
)


def test_any_instance_of():

    assert "" == AnyInstanceOf(str)
    assert 99 != AnyInstanceOf(str)

    class Foo(object):
        pass
    foo = Foo()

    assert foo == AnyInstanceOf(Foo)
    assert foo == AnyInstanceOf(object)
    assert foo != AnyInstanceOf(foo)

    assert repr(AnyInstanceOf(str)) == "<AnyInstanceOf-str>"
    assert repr(AnyInstanceOf(Foo)) == "<AnyInstanceOf-Foo>"

    assert AnyInstanceOf == AnyInstanceOf(type)
    assert str == AnyInstanceOf(type)
    assert int == AnyInstanceOf(type)
    assert type == AnyInstanceOf(type)


def test_wait_for_call():
    mock = Mock()

    def call_after(seconds):
        eventlet.sleep(seconds)
        mock.method()

    # should not raise
    eventlet.spawn(call_after, 0)
    with wait_for_call(1, mock.method):
        pass

    mock.reset_mock()

    with pytest.raises(eventlet.Timeout):
        eventlet.spawn(call_after, 1)
        with wait_for_call(0, mock.method):
            pass  # pragma: no cover


@pytest.mark.usefixtures("rabbit_config")
def test_get_extension():

    class Service(object):
        name = "service"

        @rpc
        def foo(self):
            pass  # pragma: no cover

        @rpc
        def bar(self):
            pass  # pragma: no cover

    container = ServiceContainer(Service)

    rpc_consumer = get_extension(container, RpcConsumer)
    foo_rpc = get_extension(container, Rpc, method_name="foo")
    bar_rpc = get_extension(container, Rpc, method_name="bar")

    extensions = container.extensions
    assert extensions == set([rpc_consumer, foo_rpc, bar_rpc])


@pytest.mark.usefixtures("rabbit_config")
def test_get_container(runner_factory):

    class ServiceX(object):
        name = "service_x"

    class ServiceY(object):
        name = "service_y"

    runner = runner_factory(ServiceX, ServiceY)

    assert get_container(runner, ServiceX).service_cls is ServiceX
    assert get_container(runner, ServiceY).service_cls is ServiceY
    assert get_container(runner, object) is None


def test_get_rabbit_connections():

    vhost = "vhost_name"

    connections = [{
        'vhost': vhost,
        'key': 'value'
    }, {
        'vhost': 'unlikely_vhost_name',
        'key': 'value'
    }]

    rabbit_manager = Mock()

    rabbit_manager.get_connections.return_value = connections
    vhost_conns = [connections[0]]
    assert get_rabbit_connections(vhost, rabbit_manager) == vhost_conns

    rabbit_manager.get_connections.return_value = None
    assert get_rabbit_connections(vhost, rabbit_manager) == []


def test_reset_rabbit_connections():

    with patch('nameko.testing.utils.get_rabbit_connections') as connections:
        connections.return_value = [{
            'vhost': 'vhost',
            'name': 'connection_name'
        }]

        rabbit_manager = Mock()
        reset_rabbit_connections('vhost', rabbit_manager)

        assert rabbit_manager.delete_connection.call_args_list == [
            call("connection_name")]


def test_reset_rabbit_connection_errors():

    rabbit_manager = Mock()

    with patch('nameko.testing.utils.get_rabbit_connections') as connections:
        connections.return_value = [{
            'vhost': 'vhost_name',
            'name': 'connection_name'
        }]

        # 500 error
        response_500 = Response()
        response_500.status_code = 500
        error_500 = HTTPError(response=response_500)
        rabbit_manager.delete_connection.side_effect = error_500

        with pytest.raises(HTTPError):
            reset_rabbit_connections("vhost_name", rabbit_manager)

        # 404 error
        response_404 = Response()
        response_404.status_code = 404
        error_404 = HTTPError(response=response_404)
        rabbit_manager.delete_connection.side_effect = error_404

        # does not raise
        reset_rabbit_connections("vhost_name", rabbit_manager)


def test_rabbit_connection_refused_error():

    # port 4 is an official unassigned port, so no one should be using it
    bad_port_uri = 'http://localhost:4'
    with pytest.raises(Exception) as exc_info:
        Client(bad_port_uri)

    message = str(exc_info.value)
    assert 'Connection error' in message


@patch('nameko.testing.utils.socket')
class TestFindFreePort(object):

    def test_default_host(self, patched_socket):
        host = '127.0.0.1'
        free_port = 9999
        mock_sock = patched_socket.socket()
        mock_sock.getsockname.return_value = [host, free_port]

        assert find_free_port() == free_port
        assert mock_sock.bind.call_args_list == [call((host, 0))]
        assert mock_sock.close.called

    def test_specified_host(self, patched_socket):
        host = '10.0.0.1'
        free_port = 9999
        mock_sock = patched_socket.socket()
        mock_sock.getsockname.return_value = [host, free_port]

        assert find_free_port(host) == free_port
        assert mock_sock.bind.call_args_list == [call((host, 0))]
        assert mock_sock.close.called


class TestResourcePipeline(object):

    @pytest.mark.parametrize('size', [1, 5, 1000])
    def test_pipeline(self, size):
        created = []
        destroyed = []

        counter = itertools.count()

        def create():
            obj = next(counter)
            created.append(obj)
            return obj

        def destroy(obj):
            destroyed.append(obj)

        with ResourcePipeline(create, destroy, size).run() as pipeline:

            # check initial size
            eventlet.sleep()  # let pipeline fill up
            # when full, created is always exactly one more than `size`
            assert pipeline.ready.qsize() == size
            assert len(created) == size + 1

            # get an item
            with pipeline.get() as item:
                # let pipeline process
                eventlet.sleep()
                # expect pipeline to have created another item
                assert pipeline.ready.qsize() == size
                assert len(created) == size + 2

            # after putting the item back
            # let pipeline process
            eventlet.sleep()
            # expect item to have been destroyed
            assert pipeline.trash.qsize() == 0
            assert destroyed == [item]

        # after shutdown (no need to yield because shutdown is blocking)
        # expect all created items to have been destroyed
        assert created == destroyed == list(range(size + 2))

    def test_create_shutdown_race(self):
        """
        Test the race condition where the pipeline shuts down while
        `create` is still executing.
        """
        created = []
        destroyed = []

        counter = itertools.count()
        creating = Event()

        def create():
            creating.send(True)
            eventlet.sleep()
            obj = next(counter)
            created.append(obj)
            return obj

        def destroy(obj):
            destroyed.append(obj)

        with ResourcePipeline(create, destroy).run():
            creating.wait()
            assert created == []

        assert created == destroyed == list(range(1))

    def test_shutdown_immediately(self):

        created = []
        destroyed = []

        counter = itertools.count()

        def create():  # pragma: no cover
            obj = next(counter)
            created.append(obj)
            return obj

        def destroy(obj):  # pragma: no cover
            destroyed.append(obj)

        with ResourcePipeline(create, destroy).run():
            pass

        assert created == destroyed == []

    def test_zero_size(self):

        with pytest.raises(RuntimeError):
            ResourcePipeline(None, None, 0)
